"""cenclave.core.sgx_docker module."""

from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional, Tuple
from uuid import UUID

from pydantic import BaseModel


class SgxDockerConfig(BaseModel):
    """Definition of container running in SGX enclave."""

    size: int
    host: str
    port: int
    app_id: UUID
    subject: str
    subject_alternative_name: str
    expiration_date: int
    client_certificate: Optional[str]
    app_dir: Path
    application: str
    healthcheck: str
    signer_key: Path

    signer_key_mountpoint: ClassVar[str] = "/root/.config/gramine/enclave-key.pem"
    app_mountpoint: ClassVar[str] = "/opt/input"
    docker_label: ClassVar[str] = "cenclave"
    entrypoint: ClassVar[str] = "cenclave-run"

    def cmd(self) -> List[str]:
        """Serialize the docker command args."""
        args = [
            "--size",
            f"{self.size}M",
            "--subject",
            self.subject,
            "--san",
            self.subject_alternative_name,
            "--id",
            str(self.app_id),
            "--application",
            self.application,
            "--expiration",
            str(self.expiration_date),
        ]

        if client_certificate := self.client_certificate:
            args.append("--client-certificate")
            args.append(client_certificate)

        return args

    def ports(self) -> Dict[str, Tuple[str, str]]:
        """Define the docker ports."""
        return {"443/tcp": (self.host, str(self.port))}

    def labels(self) -> Dict[str, str]:
        """Define the docker labels."""
        return {
            SgxDockerConfig.docker_label: "1",
            "healthcheck_endpoint": self.healthcheck,
        }

    def volumes(self) -> Dict[str, Dict[str, str]]:
        """Define the docker volumes."""
        return {
            f"{self.app_dir.resolve()}": {
                "bind": SgxDockerConfig.app_mountpoint,
                "mode": "rw",
            },
            "/var/run/aesmd": {"bind": "/var/run/aesmd", "mode": "rw"},
            f"{self.signer_key.resolve()}": {
                "bind": SgxDockerConfig.signer_key_mountpoint,
                "mode": "rw",
            },
        }

    @staticmethod
    def devices() -> List[str]:
        """Define the docker devices."""
        return [
            "/dev/sgx_enclave:/dev/sgx_enclave:rw",
            "/dev/sgx_provision:/dev/sgx_provision:rw",
            "/dev/sgx/enclave:/dev/sgx/enclave:rw",
            "/dev/sgx/provision:/dev/sgx/provision:rw",
        ]

    @staticmethod
    def load(docker_attrs: Dict[str, Any], docker_labels: Any):
        """Load the docker configuration from the container."""
        data_map: Dict[str, Any] = {}

        cmd = docker_attrs["Config"]["Cmd"]
        port = docker_attrs["HostConfig"]["PortBindings"]
        signer_key = next(
            filter(
                lambda mount: mount["Destination"]
                == SgxDockerConfig.signer_key_mountpoint,
                docker_attrs["Mounts"],
            )
        )
        app = next(
            filter(
                lambda mount: mount["Destination"] == SgxDockerConfig.app_mountpoint,
                docker_attrs["Mounts"],
            )
        )

        i = 0
        while i < len(cmd):
            key = cmd[i][2:]
            if i + 1 == len(cmd):
                data_map[key] = True
                i += 1
                break

            if cmd[i + 1].startswith("--"):
                data_map[key] = True
                i += 1
                continue

            data_map[key] = cmd[i + 1]
            i += 2

        return SgxDockerConfig(
            size=int(data_map["size"][:-1]),
            host=port["443/tcp"][0]["HostIp"],
            subject=data_map["subject"],
            subject_alternative_name=data_map["san"],
            app_id=UUID(data_map["id"]),
            expiration_date=int(data_map["expiration"]),
            client_certificate=data_map.get("client-certificate"),
            app_dir=Path(app["Source"]),
            application=data_map["application"],
            port=int(port["443/tcp"][0]["HostPort"]),
            healthcheck=docker_labels["healthcheck_endpoint"],
            signer_key=Path(signer_key["Source"]),
        )



"""Test helpers functions."""

from pathlib import Path

from cenclave.command.sgx_operator.evidence import guess_pccs_url


def test_guess_pccs_url():
    """Test guess_pccs_url."""
    conf = Path(__file__).parent / "data/sgx_default_qcnl.conf"

    assert guess_pccs_url(aemsd_conf_file=conf) == "https://example.cosmian.com"



"""cenclave.command.code_provider.unseal module."""

import sys
from pathlib import Path

from cenclave_lib_crypto.seal_box import unseal

from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser("unseal", help="unseal file using NaCl's Seal Box")

    parser.add_argument(
        "--input",
        type=Path,
        metavar="FILE",
        required=True,
        help="path to the file to unseal",
    )

    parser.add_argument(
        "--private-key",
        type=Path,
        metavar="FILE",
        required=True,
        help="path to raw X25519 private key",
    )

    parser.add_argument(
        "--output",
        type=Path,
        metavar="FILE",
        help="path to write the file unsealed",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    LOG.info("Unsealing %s...", args.input)

    private_key: bytes = args.private_key.read_bytes()
    encrypted_data: bytes = args.input.read_bytes()

    data: bytes = unseal(encrypted_data, private_key)

    if args.output:
        args.output.write_bytes(data)
        LOG.info("File successfully unsealed to %s", args.output)
    else:
        LOG.info("Data sucessfully unsealed!")
        LOG.info(
            "----------------------------------------------------------------------"
        )
        sys.stdout.buffer.write(data)



"""cenclave.main module."""

import argparse
import os
import traceback
from warnings import filterwarnings  # noqa: E402

filterwarnings("ignore")  # noqa: E402

# pylint: disable=wrong-import-position
import cenclave
from cenclave.command.code_provider import (
    decrypt,
    encrypt,
    keygen,
    localtest,
    package,
    scaffold,
    seal,
    unseal,
    verify,
)
from cenclave.command.sgx_operator import (
    evidence,
    list_all,
    logs,
    restart,
    run,
    spawn,
    status,
    stop,
    test,
)
from cenclave.log import LOGGER as LOG
from cenclave.log import setup_logging


# pylint: disable=too-many-statements
def main() -> int:
    """Entrypoint of the CLI."""
    parser = argparse.ArgumentParser(
        description="Cosmian Enclave CLI" f" - {cenclave.__version__}"
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"{cenclave.__version__}",
        help="version of %(prog)s CLI",
    )

    subparsers = parser.add_subparsers(title="commands")

    keygen.add_subparser(subparsers)
    encrypt.add_subparser(subparsers)
    decrypt.add_subparser(subparsers)
    evidence.add_subparser(subparsers)
    scaffold.add_subparser(subparsers)
    list_all.add_subparser(subparsers)
    logs.add_subparser(subparsers)
    package.add_subparser(subparsers)
    restart.add_subparser(subparsers)
    run.add_subparser(subparsers)
    status.add_subparser(subparsers)
    seal.add_subparser(subparsers)
    unseal.add_subparser(subparsers)
    spawn.add_subparser(subparsers)
    stop.add_subparser(subparsers)
    test.add_subparser(subparsers)
    localtest.add_subparser(subparsers)
    verify.add_subparser(subparsers)

    args = parser.parse_args()

    setup_logging(False)

    try:
        func = args.func
    except AttributeError:
        parser.error("too few arguments")
        return 1

    try:
        func(args)
        return 0
    # pylint: disable=broad-except
    except Exception as e:
        if os.getenv("BACKTRACE") == "full":
            traceback.print_exc()

        LOG.error(e)
        return 1


if __name__ == "__main__":
    main()



"""Test model/args.py."""

from pathlib import Path

from cenclave.core.test_docker import TestDockerConfig


def test_volumes():
    """Test `volumes` function."""
    ref_conf = TestDockerConfig(
        port=5000,
        code="/tmp/code.tar",
        application="app:app",
        sealed_secrets=None,
        secrets=None,
    )

    assert ref_conf.volumes() == {
        "/tmp/code.tar": {
            "bind": "/app",
            "mode": "rw",
        }
    }

    ref_conf = TestDockerConfig(
        port=5000,
        code=Path("/tmp/code.tar"),
        application="app:app",
        sealed_secrets=Path("/tmp/sealed_secrets"),
        secrets=Path("/tmp/secrets.json"),
    )

    assert ref_conf.volumes() == {
        "/tmp/code.tar": {
            "bind": "/app",
            "mode": "rw",
        },
        "/tmp/secrets.json": {
            "bind": "/root/.cache/cenclave/secrets.json",
            "mode": "rw",
        },
        "/tmp/sealed_secrets": {
            "bind": "/root/.cache/cenclave/sealed_secrets.json",
            "mode": "rw",
        },
    }


def test_cmd():
    """Test `cmd` function."""
    ref_conf = TestDockerConfig(
        port=5000,
        code="/tmp/code.tar",
        application="app:app",
        sealed_secrets=None,
        secrets=None,
    )

    assert ref_conf.cmd() == [
        "--application",
        "app:app",
        "--debug",
    ]

    ref_conf = TestDockerConfig(
        port=5000,
        code=Path("/tmp/code.tar"),
        application="app:app",
        sealed_secrets=Path("/tmp/sealed_secrets"),
        secrets=Path("/tmp/secrets.json"),
    )

    assert ref_conf.cmd() == [
        "--application",
        "app:app",
        "--debug",
    ]


def test_ports():
    """Test `ports` function."""
    ref_conf = TestDockerConfig(
        port=5000,
        code="/tmp/code.tar",
        application="app:app",
        sealed_secrets=None,
        secrets=None,
    )

    assert ref_conf.ports() == {"5000/tcp": ("127.0.0.1", "5000")}

    ref_conf = TestDockerConfig(
        port=5000,
        code=Path("/tmp/code.tar"),
        application="app:app",
        sealed_secrets=Path("/tmp/sealed_secrets"),
        secrets=Path("/tmp/secrets.json"),
    )

    assert ref_conf.ports() == {"5000/tcp": ("127.0.0.1", "5000")}



"""cenclave.core.bootstrap module."""

from typing import Any, Callable, Dict, Iterable, Optional, Union
from uuid import UUID

import requests
import urllib3
from pydantic import BaseModel

from cenclave.core.base64 import base64url_encode
from cenclave.core.clock_tick import ClockTick
from cenclave.error import UnexpectedResponse


class ConfigurationPayload(BaseModel):
    """Definition of the bootstrap server payload."""

    app_id: UUID
    secrets: Optional[Any]
    sealed_secrets: Optional[bytes]
    code_secret_key: Optional[bytes]
    ssl_private_key: Optional[str]

    def payload(self) -> Dict[str, Any]:
        """Build the payload to send to the configuration server."""
        data: Dict[str, Any] = {
            "uuid": str(self.app_id),
        }

        if self.secrets:
            data["app_secrets"] = self.secrets

        if self.sealed_secrets:
            data["app_sealed_secrets"] = base64url_encode(self.sealed_secrets)

        if self.code_secret_key:
            data["code_secret_key"] = self.code_secret_key.hex()

        if self.ssl_private_key:
            data["ssl_private_key"] = self.ssl_private_key

        return data


def configure_app(url: str, data: Dict[str, Any], verify: Union[bool, str] = True):
    """Send the secrets to the configuration server."""
    r = requests.post(
        url=url,
        json=data,
        headers={"Content-Type": "application/json"},
        verify=verify,
        timeout=60,
    )

    if not r.ok:
        raise UnexpectedResponse(
            "Fail to send data to the configuration server "
            f"(Response {r.status_code} {r.text})"
        )


def wait_for_conf_server(
    clock: ClockTick,
    url: str,
    verify: Union[bool, str] = True,
    extra_check: Optional[Callable] = None,
    extra_check_args: Iterable = (),
):
    """Hold on until the configuration server is up and listing."""
    while clock.tick():
        if extra_check:
            extra_check(*extra_check_args)

        if is_waiting_for_secrets(url, verify):
            break


def is_waiting_for_secrets(url: str, verify: Union[bool, str] = True) -> bool:
    """Check whether the configuration server is up."""
    try:
        response = requests.get(url=url, verify=verify, timeout=5)

        if response.status_code == 200 and "Mse-Status" in response.headers:
            return True
    except requests.exceptions.Timeout:
        return False
    except requests.exceptions.SSLError:
        return False
    except requests.exceptions.ConnectionError:
        return False

    return False


# pylint: disable=too-many-positional-arguments
def wait_for_app_server(
    clock: ClockTick,
    url: str,
    healthcheck_endpoint: str,
    verify: Union[bool, str] = True,
    extra_check: Optional[Callable] = None,
    extra_check_args: Iterable = (),
):
    """Hold on until the configuration server is stopped and the app starts."""
    while clock.tick():
        if extra_check:
            extra_check(*extra_check_args)

        if is_ready(url, healthcheck_endpoint, verify):
            break


def is_ready(
    url: str, healthcheck_endpoint: str, verify: Union[bool, str] = True
) -> bool:
    """Check whether the app server is up."""
    try:
        response = requests.get(
            url=f"{url}{healthcheck_endpoint}",
            verify=verify,
            timeout=5,
        )

        if response.status_code != 503 and "Mse-Status" not in response.headers:
            return True
    except (
        requests.exceptions.Timeout,
        requests.exceptions.SSLError,
    ):
        return False
    except requests.exceptions.ConnectionError as exc:
        err, *_ = exc.args
        if isinstance(err, urllib3.exceptions.ProtocolError):
            return True
        return False

    return False



"""cenclave.command.helpers module."""

import socket
from pathlib import Path
from typing import Optional

from docker import from_env
from docker.client import DockerClient
from docker.errors import DockerException, NotFound
from docker.models.containers import Container

from cenclave.error import AppContainerNotFound, AppContainerNotRunning
from cenclave.log import LOGGER as LOG


def get_client_docker() -> DockerClient:
    """Create a Docker client or exit if daemon is down."""
    try:
        return from_env()
    except DockerException as exc:
        LOG.warning("Docker seems not running. Please enable Docker daemon.")
        LOG.info("Docker is needed to manage your app.")
        LOG.info("Please refer to the documentation for more details.")
        raise exc


def app_container_exists(client: DockerClient, name: str) -> Optional[Container]:
    """Check whether a container exists based on its `name`."""
    try:
        return get_app_container(client, name)
    except AppContainerNotFound:
        return None


def get_app_container(client: DockerClient, name: str) -> Container:
    """Get a container based on its `name`."""
    try:
        return client.containers.get(name)
    except NotFound as exc:
        raise AppContainerNotFound(
            f"Can't find the container for application '{name}'"
        ) from exc


def is_running(container) -> bool:
    """Test whether the given container is running."""
    return container.status == "running"


def get_running_app_container(client: DockerClient, name: str) -> Container:
    """Raise an error if the container is not running or return the running container."""
    container = get_app_container(client, name)

    if not is_running(container):
        raise AppContainerNotRunning(
            f"Your application '{name}' is not running. "
            "Run `cenclave logs` for more details"
        )

    return container


def load_docker_image(client: DockerClient, image_tar_path: Path) -> str:
    """Load the docker image from the image tarball."""
    LOG.info("Loading the docker image...")
    with open(image_tar_path, "rb") as f:
        image = client.images.load(f.read())
        return image[0].tags[0]


def is_port_free(port: int):
    """Check whether a given `port` is free."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.bind(("", port))
        sock.close()
    except OSError:
        return False

    return True


def enclave_size_integer(n: str) -> int:
    """Define a new integer type for the enclave size arg."""
    m = int(n)
    min_value = 1024
    if m < min_value:
        raise ValueError(f"Enclave size should be greater than {min_value}")

    is_power_of_two = (m & (m - 1) == 0) and m != 0
    if not is_power_of_two:
        raise ValueError("Enclave size should be a power of two (lower than EPC size)")

    return m



"""cenclave.model.evidence module."""

import base64
import json
from pathlib import Path
from typing import Any, Dict, Tuple

from cryptography.hazmat.primitives.asymmetric.types import PublicKeyTypes
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PublicFormat,
    load_pem_public_key,
)
from cryptography.x509 import (
    Certificate,
    CertificateRevocationList,
    load_pem_x509_certificate,
    load_pem_x509_crl,
)
from pydantic import BaseModel

from cenclave.core.no_sgx_docker import NoSgxDockerConfig


class ApplicationEvidence(BaseModel):
    """Definition of an enclave evidence."""

    ratls_certificate: Certificate
    root_ca_crl: CertificateRevocationList
    pck_platform_crl: CertificateRevocationList
    tcb_info: bytes
    qe_identity: bytes
    tcb_cert: Certificate
    signer_pk: PublicKeyTypes
    input_args: NoSgxDockerConfig

    class Config:
        """Overwrite internal structure."""

        arbitrary_types_allowed = True

    @property
    def collaterals(
        self,
    ) -> Tuple[
        bytes, bytes, Certificate, CertificateRevocationList, CertificateRevocationList
    ]:
        """Return the PCCS collaterals."""
        return (
            self.tcb_info,
            self.qe_identity,
            self.tcb_cert,
            self.root_ca_crl,
            self.pck_platform_crl,
        )

    @staticmethod
    def load(path: Path):
        """Load the evidence from a json file."""
        with open(path, encoding="utf8") as f:
            data_map = json.load(f)

            return ApplicationEvidence(
                input_args=NoSgxDockerConfig(**data_map["input_args"]),
                ratls_certificate=load_pem_x509_certificate(
                    data_map["ratls_certificate"].encode("utf-8")
                ),
                root_ca_crl=load_pem_x509_crl(data_map["root_ca_crl"].encode("utf-8")),
                pck_platform_crl=load_pem_x509_crl(
                    data_map["pck_platform_crl"].encode("utf-8")
                ),
                tcb_info=base64.b64decode(data_map["tcb_info"].encode("utf-8")),
                qe_identity=base64.b64decode(data_map["qe_identity"].encode("utf-8")),
                tcb_cert=load_pem_x509_certificate(
                    data_map["tcb_cert"].encode("utf-8")
                ),
                signer_pk=load_pem_public_key(
                    data_map["signer_pk"].encode("utf-8"),
                ),
            )

    def save(self, path: Path) -> None:
        """Save the evidence into a json file."""
        with open(path, "w", encoding="utf8") as f:
            data_map: Dict[str, Any] = {
                "input_args": {
                    "subject": self.input_args.subject,
                    "subject_alternative_name": self.input_args.subject_alternative_name,
                    "expiration_date": (
                        self.input_args.expiration_date
                        if self.input_args.expiration_date
                        else None
                    ),
                    "size": self.input_args.size,
                    "app_id": str(self.input_args.app_id),
                    "application": self.input_args.application,
                },
                "ratls_certificate": self.ratls_certificate.public_bytes(
                    encoding=Encoding.PEM
                ).decode("utf-8"),
                "root_ca_crl": self.root_ca_crl.public_bytes(
                    encoding=Encoding.PEM,
                ).decode("utf-8"),
                "pck_platform_crl": self.pck_platform_crl.public_bytes(
                    encoding=Encoding.PEM,
                ).decode("utf-8"),
                "tcb_info": base64.b64encode(self.tcb_info).decode("utf-8"),
                "qe_identity": base64.b64encode(self.qe_identity).decode("utf-8"),
                "tcb_cert": self.tcb_cert.public_bytes(encoding=Encoding.PEM).decode(
                    "utf-8"
                ),
                "signer_pk": self.signer_pk.public_bytes(
                    encoding=Encoding.PEM,
                    format=PublicFormat.SubjectPublicKeyInfo,
                ).decode("utf-8"),
            }

            json.dump(data_map, f, indent=4)



"""cenclave.core.conf module."""

from __future__ import annotations

import os
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import toml
from pydantic import BaseModel, constr, validator

from cenclave.error import BadApplicationInput

if TYPE_CHECKING:
    Str255 = str
    Str16 = str
    StrUnlimited = str
else:
    Str255 = constr(min_length=1, max_length=255, strip_whitespace=True)
    Str16 = constr(min_length=1, max_length=16, strip_whitespace=True)
    StrUnlimited = constr(min_length=1)


def absolute_from_conf_file(conf_file: Path, path: Path) -> Path:
    """Make the `path` absolute from `conf_file.parent`."""
    if not path.is_absolute():
        return (conf_file.parent / path).resolve()

    return path


class AppConfParsingOption(Enum):
    """Parsing option for AppConf."""

    All = 0
    SkipCloud = 1
    UseInsecureCloud = 2


class SSLConf(BaseModel):
    """Definition of the app owner certificate."""

    # The domain name of the app
    domain_name: Str255
    # The path to the ssl private key
    private_key: Path
    # The path to the ssl certificate chain
    certificate: Path

    @property
    def certificate_data(self) -> str:
        """Get the data from certificate file."""
        return self.certificate.read_text()

    @property
    def private_key_data(self) -> str:
        """Get the data from private_key file."""
        return self.private_key.read_text()


class AppConf(BaseModel):
    """Define the application configuration."""

    # Name of the enclave instance
    name: Str255
    # from python_flask_module import python_flask_variable_name
    python_application: Str255
    # Endpoint to use to check if the application is up and sane
    healthcheck_endpoint: Str255
    # The command to test the application
    tests_cmd: str
    # The package to install before testing the application
    tests_requirements: List[str]

    @validator("healthcheck_endpoint", pre=False)
    # pylint: disable=no-self-argument,unused-argument
    def check_healthcheck_endpoint(cls, v: str):
        """Validate that `healthcheck_endpoint` is an endpoint."""
        if v.startswith("/"):
            return v
        raise ValueError('healthcheck_endpoint should start with a "/"')

    # pylint: disable=unused-argument
    @staticmethod
    def load(
        path: Optional[Path],
        option: AppConfParsingOption = AppConfParsingOption.All,
    ):
        """Load the configuration from a toml file."""
        path = path.expanduser() if path else Path(os.getcwd()) / "config.toml"

        with open(path, encoding="utf8") as f:
            data_map = toml.load(f)
            app = AppConf(**data_map)

            return app

    def save(self, path: Path) -> None:
        """Save the configuration into a toml file."""
        with open(path, "w", encoding="utf8") as f:
            data_map: Dict[str, Any] = {
                "name": self.name,
                "python_application": self.python_application,
                "healthcheck_endpoint": self.healthcheck_endpoint,
                "tests_cmd": self.tests_cmd,
                "tests_requirements": self.tests_requirements,
            }

            toml.dump(data_map, f)

    @property
    def python_module(self):
        """Get the python module from python_application."""
        split_str = self.python_application.split(":")
        if len(split_str) != 2:
            raise BadApplicationInput(
                "`python_application` is malformed. Expected format: `module:variable`"
            )
        return split_str[0]

    @property
    def python_variable(self):
        """Get the python variable from python_application."""
        split_str = self.python_application.split(":")
        if len(split_str) != 2:
            raise BadApplicationInput(
                "`python_application` is malformed. Expected format: `module:variable`"
            )
        return split_str[1]



"""cenclave.command.code_provider.seal module."""

import sys
from pathlib import Path

from cenclave_lib_crypto.seal_box import seal
from intel_sgx_ra.quote import Quote
from intel_sgx_ra.ratls import ratls_verify

from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser(
        "seal",
        help="seal file using NaCl sealed box. "
        "Recipient is either raw X25519 public key or "
        "extracted from RA-TLS certificate with enclave's "
        "public key in REPORT_DATA field of SGX quote",
    )

    parser.add_argument(
        "--input",
        type=Path,
        metavar="FILE",
        required=True,
        help="path to the file to seal",
    )

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "--receiver-enclave",
        type=Path,
        metavar="FILE",
        help="path to RA-TLS certificate of the enclave",
    )

    group.add_argument(
        "--receiver-public-key",
        type=Path,
        metavar="FILE",
        help="path to raw X25519 public key",
    )

    parser.add_argument(
        "--output",
        type=Path,
        metavar="FILE",
        help="path to write the file sealed",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    LOG.info("Sealing %s...", args.input)

    enclave_pk: bytes
    if args.receiver_enclave:
        quote: Quote = ratls_verify(args.receiver_enclave)
        enclave_pk = quote.report_body.report_data[32:64]
    else:
        enclave_pk = args.receiver_public_key.read_bytes()

    encrypted_data: bytes = seal(args.input.read_bytes(), enclave_pk)

    if args.output:
        args.output.write_bytes(encrypted_data)
        LOG.info("File sealed to %s", args.output)
    else:
        LOG.info("Data sucessfully sealed!")
        LOG.info(
            "----------------------------------------------------------------------"
        )
        sys.stdout.buffer.write(encrypted_data)



"""cenclave.model.package module."""

import tarfile
from pathlib import Path

from pydantic import BaseModel

from cenclave.error import PackageMalformed

DEFAULT_CODE_DIR = "src"
DEFAULT_CONFIG_FILENAME = "config.toml"
DEFAULT_TEST_DIR = "tests"
DEFAULT_DOCKERFILE_FILENAME = "Dockerfile"
DEFAULT_SECRETS_FILENAME = "secrets.json"
DEFAULT_SEAL_SECRETS_FILENAME = "secrets_to_seal.json"

CODE_TAR_NAME = "app.tar"
DOCKER_IMAGE_TAR_NAME = "image.tar"
CONFIG_NAME = "config.toml"
TEST_TAR_NAME = "tests.tar"


class CodePackage(BaseModel):
    """Definition of a code package."""

    code_tar: Path
    image_tar: Path
    test_tar: Path
    config_path: Path

    def create(
        self,
        output_tar: Path,
    ):
        """Create the package containing the code and Docker image tarballs."""
        with tarfile.open(output_tar, "w:") as tar_file:
            tar_file.add(self.code_tar, CODE_TAR_NAME)
            tar_file.add(self.image_tar, DOCKER_IMAGE_TAR_NAME)
            tar_file.add(self.test_tar, TEST_TAR_NAME)
            tar_file.add(self.config_path, CONFIG_NAME)

    @staticmethod
    def extract(workspace: Path, package: Path):
        """Extract the code and image from the tarball."""
        with tarfile.open(package, "r") as f:
            f.extractall(path=workspace)

        code_tar_path = workspace / CODE_TAR_NAME
        image_tar_path = workspace / DOCKER_IMAGE_TAR_NAME
        code_config_path = workspace / CONFIG_NAME
        test_tar_path = workspace / TEST_TAR_NAME

        if not code_tar_path.exists():
            raise PackageMalformed(
                f"'{CODE_TAR_NAME}' was not found in the cenclave package"
            )

        if not image_tar_path.exists():
            raise PackageMalformed(
                f"'{DOCKER_IMAGE_TAR_NAME}' was not found in the cenclave package"
            )

        if not test_tar_path.exists():
            raise PackageMalformed(
                f"'{TEST_TAR_NAME}' was not found in the cenclave package"
            )

        if not code_config_path.exists():
            raise PackageMalformed(
                f"'{CONFIG_NAME}' was not found in the cenclave package"
            )

        return CodePackage(
            code_tar=code_tar_path,
            image_tar=image_tar_path,
            test_tar=test_tar_path,
            config_path=code_config_path,
        )



"""cenclave.core.clock_tick module."""

import time

from cenclave.error import Timeout


class ClockTick:
    """Class to monitor the spent time."""

    def __init__(self, period: int, timeout: int, message: str):
        """Initialize the clock."""
        self.elapsed = 0
        self.timeout = timeout
        self.period = period
        self.message = message

    def tick(self) -> bool:
        """Start ticking."""
        if self.elapsed > self.timeout:
            raise Timeout(self.message)

        time.sleep(self.period)
        self.elapsed += self.period

        return True



"""cenclave.core.enclave module."""

import logging
import re
import uuid
from pathlib import Path
from typing import Optional, Tuple, Union

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.x509 import Certificate, CertificateRevocationList
from docker.client import DockerClient
from docker.errors import NotFound
from intel_sgx_ra.attest import verify_quote
from intel_sgx_ra.maa.attest import verify_quote as azure_verify_quote
from intel_sgx_ra.ratls import ratls_verify
from intel_sgx_ra.signer import mr_signer_from_pk

from cenclave.core.no_sgx_docker import NoSgxDockerConfig
from cenclave.error import (
    AppContainerError,
    RatlsVerificationFailure,
    WrongMREnclave,
    WrongMRSigner,
)


def compute_mr_enclave(
    client: DockerClient,
    image: str,
    app_args: NoSgxDockerConfig,
    app_path: Path,
    docker_path_log: Path,
) -> str:
    """Compute the MR enclave."""
    container_name = str(uuid.uuid4())
    output = b""

    try:
        _ = client.containers.run(
            image,
            name=container_name,
            command=app_args.cmd(),
            volumes=app_args.volumes(app_path),
            entrypoint=NoSgxDockerConfig.entrypoint,
            # We do not remove the container to be able to print the error (if some)
            remove=False,
            detach=False,
            stdout=True,
            stderr=True,
        )
    except Exception as exc:
        raise AppContainerError(
            f"Error starting the docker (see logs at {docker_path_log})"
        ) from exc
    finally:
        try:
            container = client.containers.get(container_name)
            # Save the docker output
            output = container.logs()
            docker_path_log.write_bytes(output)
            container.stop(timeout=1)
            # We need to remove the container since we declare remove=False previously
            container.remove()

        except NotFound:
            pass

    # Get the mr_enclave from the docker output
    pattern = "Measurement:\n[ ]*([a-z0-9]{64})"
    m = re.search(pattern.encode("utf-8"), output)

    if not m:
        raise RatlsVerificationFailure(
            f"Fail to compute mr_enclave! See {docker_path_log} for more details."
        )

    return str(m.group(1).decode("utf-8"))


def verify_enclave(
    signer_pk: Union[RSAPublicKey, Path, bytes],
    ratls_certificate: Union[str, bytes, Path, Certificate],
    fingerprint: Optional[str],
    collaterals: Optional[
        Tuple[
            bytes,
            bytes,
            Certificate,
            CertificateRevocationList,
            CertificateRevocationList,
        ]
    ] = None,
    pccs_url: Optional[str] = None,
):
    """Verify an enclave trustworthiness."""
    # Compute MRSIGNER value from public key
    mrsigner = mr_signer_from_pk(signer_pk)

    # Check certificate's public key in quote's user report data
    quote = ratls_verify(ratls_certificate)

    # Check MRSIGNER
    if quote.report_body.mr_signer != mrsigner:
        raise WrongMRSigner(
            "Enclave signer is wrong "
            f"(read {bytes(quote.report_body.mr_signer).hex()} "
            f"but should be {bytes(mrsigner).hex()})"
        )

    logging.info("MRSIGNER: %s", quote.report_body.mr_signer.hex())
    logging.info("MRENCLAVE: %s", quote.report_body.mr_enclave.hex())

    if collaterals is None and pccs_url is None:
        # Azure DCAP attestation through MAA service
        azure_verify_quote(quote=quote)
    else:
        # Intel DCAP attestation using PCCS url or directly collaterals if provided
        verify_quote(quote=quote, collaterals=collaterals, pccs_url=pccs_url)

    # Check MRENCLAVE
    if fingerprint:
        if quote.report_body.mr_enclave != bytes.fromhex(fingerprint):
            raise WrongMREnclave(
                "Code fingerprint is wrong "
                f"(read {bytes(quote.report_body.mr_enclave).hex()} "
                f"but should be {fingerprint})"
            )



"""cenclave.core.fs module."""

import tarfile
from pathlib import Path
from typing import Iterator, List


def whitelist() -> List[str]:
    """Get a default whitelist."""
    return list({"requirements.txt"})


def is_hidden(path: Path) -> bool:
    """Check if `path` contains parts with hidden file or folder.

    Parameters
    ----------
    path : Path
        Path to check if any part is hidden.

    Returns
    -------
    bool
        True if any part of `path` is hidden, False otherwise.

    """
    return any((part.startswith(".") for part in path.parts))


def ls(dir_path: Path, dot_files: bool = False) -> Iterator[Path]:
    """Recursive listing of files `dir_path`.

    Parameters
    ----------
    dir_path : Path
        Path to the directory.
    dot_files : bool
        Whether you want to list dot files.

    Yields
    -------
    Path
        Path to a file within `dir_path`.

    """
    for path in sorted(dir_path.absolute().rglob("*")):  # type: Path
        if path.is_file():
            if not dot_files and not is_hidden(path.relative_to(dir_path)):
                yield path


def tar(dir_path: Path, tar_path: Path, dot_files: bool = False) -> Path:
    """Tar directory `dir_path` to `tar_path`.

    Parameters
    ----------
    dir_path : Path
        Directory path to tar.
    tar_path : Path
        Path to output the tar file created.
    dot_files : bool

    Returns
    -------
    Path
        Path to the tar file created.

    """
    with tarfile.open(tar_path, "w:") as tar_file:
        for path in ls(dir_path, dot_files):
            rel_path: Path = path.relative_to(dir_path)
            tar_file.add(path, rel_path)

    return tar_path


def untar(dir_path: Path, tar_file: Path):
    """Untar file `tar_file` to `dir_path`.

    Parameters
    ----------
    dir_path : Path
        Directory output path.
    tar_file : Path
        Path to of the tar file to untar.

    """
    with tarfile.open(tar_file, "r:") as f:
        f.extractall(dir_path)



"""Test model/args.py."""

from pathlib import Path

from cenclave.core.no_sgx_docker import NoSgxDockerConfig
from cenclave.core.sgx_docker import SgxDockerConfig


def test_from_sgx():
    """Test the `from_sgx` method."""
    ref_conf = NoSgxDockerConfig(
        subject="CN=localhost,O=Big Company,C=FR,L=Paris,ST=Ile-de-France",
        subject_alternative_name="localhost",
        expiration_date=1714058115,
        size=4096,
        app_id="63322f85-1ff8-4483-91ae-f18d7398d157",
        application="app:app",
    )

    conf = NoSgxDockerConfig.from_sgx(
        SgxDockerConfig(
            size=4096,
            host="localhost",
            port=7788,
            subject="CN=localhost,O=Big Company,C=FR,L=Paris,ST=Ile-de-France",
            subject_alternative_name="localhost",
            app_id="63322f85-1ff8-4483-91ae-f18d7398d157",
            expiration_date=1714058115,
            app_dir="/home/cosmian/workspace/sgx_operator/",
            application="app:app",
            healthcheck="/health",
            signer_key="/opt/cosmian-internal/cosmian-signer-key.pem",
        )
    )

    assert conf == ref_conf


def test_volumes():
    """Test `volumes` function."""
    ref_conf = NoSgxDockerConfig(
        subject="CN=localhost,O=Big Company,C=FR,L=Paris,ST=Ile-de-France",
        subject_alternative_name="localhost",
        expiration_date=1714058115,
        size=4096,
        app_id="63322f85-1ff8-4483-91ae-f18d7398d157",
        application="app:app",
    )

    assert ref_conf.volumes(Path("/tmp/")) == {
        "/tmp": {
            "bind": "/opt/input",
            "mode": "rw",
        }
    }


def test_cmd():
    """Test `cmd` function."""
    ref_conf = NoSgxDockerConfig(
        subject="CN=localhost,O=Big Company,C=FR,L=Paris,ST=Ile-de-France",
        subject_alternative_name="localhost",
        expiration_date=1714058115,
        size=4096,
        app_id="63322f85-1ff8-4483-91ae-f18d7398d157",
        application="app:app",
    )

    assert ref_conf.cmd() == [
        "--size",
        "4096M",
        "--subject",
        "CN=localhost,O=Big Company,C=FR,L=Paris,ST=Ile-de-France",
        "--san",
        "localhost",
        "--id",
        "63322f85-1ff8-4483-91ae-f18d7398d157",
        "--application",
        "app:app",
        "--dry-run",
        "--expiration",
        "1714058115",
    ]

    ref_conf = NoSgxDockerConfig(
        subject="CN=localhost,O=Big Company,C=FR,L=Paris,ST=Ile-de-France",
        subject_alternative_name="localhost",
        size=4096,
        app_id="63322f85-1ff8-4483-91ae-f18d7398d157",
        application="app:app",
    )

    assert ref_conf.cmd() == [
        "--size",
        "4096M",
        "--subject",
        "CN=localhost,O=Big Company,C=FR,L=Paris,ST=Ile-de-France",
        "--san",
        "localhost",
        "--id",
        "63322f85-1ff8-4483-91ae-f18d7398d157",
        "--application",
        "app:app",
        "--dry-run",
    ]



"""cenclave.command.sgx_operator.evidence module."""

import json
import socket
import ssl
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

from cryptography.hazmat.primitives.serialization import Encoding, load_pem_private_key
from cryptography.x509 import load_pem_x509_certificate
from docker.models.containers import Container
from intel_sgx_ra.attest import retrieve_collaterals
from intel_sgx_ra.ratls import get_server_certificate, ratls_verify

from cenclave.command.helpers import get_client_docker, get_running_app_container
from cenclave.core.no_sgx_docker import NoSgxDockerConfig
from cenclave.core.sgx_docker import SgxDockerConfig
from cenclave.log import LOGGER as LOG
from cenclave.model.evidence import ApplicationEvidence


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser(
        "evidence",
        help="collect the evidences to verify on offline mode "
        "the application and the enclave",
    )

    pccs_url_default = guess_pccs_url() or "https://pccs.example.com"
    parser.add_argument(
        "--pccs",
        type=str,
        help=f"URL to the PCCS (default: {pccs_url_default})",
        default=pccs_url_default,
    )

    parser.add_argument(
        "--output",
        type=Path,
        required=True,
        help="the directory to write the evidence file",
    )

    parser.add_argument(
        "name",
        type=str,
        help="the name of the application",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    if not args.output.is_dir():
        raise NotADirectoryError(f"`{args.output}` does not exist")

    client = get_client_docker()
    container = get_running_app_container(client, args.name)

    collect_evidence_and_certificate(
        container=container, pccs_url=args.pccs, output=args.output
    )


# pylint: disable=too-many-locals
def collect_evidence_and_certificate(
    container: Container,
    pccs_url: str,
    output: Path,
):
    """Collect evidence JSON file and RA-TLS certificate from running enclave."""
    LOG.info("Collecting the enclave and application evidences...")

    docker = SgxDockerConfig.load(container.attrs, container.labels)
    input_args = NoSgxDockerConfig.from_sgx(docker_config=docker)

    # Get the certificate from the application
    try:
        ratls_cert = load_pem_x509_certificate(
            get_server_certificate((docker.host, docker.port)).encode("utf-8")
        )
    except (ssl.SSLZeroReturnError, socket.gaierror, ssl.SSLEOFError) as exc:
        raise ConnectionError(
            f"Can't reach {docker.host}:{docker.port}. "
            "Are you sure the application is still running?"
        ) from exc

    quote = ratls_verify(ratls_cert)

    (
        tcb_info,
        qe_identity,
        tcb_cert,
        root_ca_crl,
        pck_platform_crl,
    ) = retrieve_collaterals(quote, pccs_url)

    signer_key = load_pem_private_key(
        docker.signer_key.read_bytes(),
        password=None,
    )

    evidence = ApplicationEvidence(
        input_args=input_args,
        ratls_certificate=ratls_cert,
        root_ca_crl=root_ca_crl,
        pck_platform_crl=pck_platform_crl,
        tcb_info=tcb_info,
        qe_identity=qe_identity,
        tcb_cert=tcb_cert,
        signer_pk=signer_key.public_key(),
    )

    evidence_path = output / "evidence.json"
    evidence.save(evidence_path)
    LOG.info("The evidence file has been generated at: %s", evidence_path)
    LOG.info("The evidence file can now be shared!")

    ratls_cert_path = output / "ratls.pem"
    ratls_cert_path.write_bytes(
        evidence.ratls_certificate.public_bytes(encoding=Encoding.PEM)
    )

    LOG.info("The RA-TLS certificate has been saved at: %s", ratls_cert_path)


def guess_pccs_url(
    aemsd_conf_file: Path = Path("/etc/sgx_default_qcnl.conf"),
) -> Optional[str]:
    """Get the pccs url from aesmd service configuration file."""
    try:
        with open(aemsd_conf_file, encoding="utf-8") as f:
            # The configuration is not a valid json: it contains comments
            # First remove them and then deserialize the json string
            content = f.readlines()
            content = [
                line
                for line in content
                if line.strip() and not line.strip().startswith("//")
            ]
            url = json.loads("".join(content)).get("pccs_url", None)

            if url:
                url = urlparse(url)
                return f"{url.scheme}://{url.netloc}"
    except Exception:  # pylint: disable=broad-except
        pass

    return None



"""cenclave.error module."""


class AppContainerNotFound(Exception):
    """Application container not found."""


class AppContainerNotRunning(Exception):
    """Application container is not running."""


class DockerBuildError(Exception):
    """Docker build failed."""


class AppContainerAlreadyRunning(Exception):
    """Application container is already running."""


class PortBusy(Exception):
    """Application port is already busy."""


class AppContainerError(Exception):
    """Application failed to start."""


class AppContainerBadState(Exception):
    """Application is not in the expecting state."""


class PackageMalformed(Exception):
    """Package is malformed."""


class Timeout(Exception):
    """Timeout occurs."""


class UnexpectedResponse(Exception):
    """Unexpected response when querying a server."""


class RatlsVerificationFailure(Exception):
    """The remote attestion failed."""


class RatlsVerificationNotSupported(Exception):
    """Proceed the remote attestion is impossible."""


class BadApplicationInput(Exception):
    """The application configuration is wrong."""


class WrongMREnclave(Exception):
    """MR enclave does not match with the expected value."""


class WrongMRSigner(Exception):
    """MR signer does not match with the expected value."""



"""cenclave.command.code_provider.verify module."""

import shutil
import tempfile
from pathlib import Path

from cryptography.hazmat.primitives.serialization import Encoding

from cenclave.command.helpers import get_client_docker, load_docker_image
from cenclave.core.enclave import compute_mr_enclave, verify_enclave
from cenclave.log import LOGGER as LOG
from cenclave.model.evidence import ApplicationEvidence
from cenclave.model.package import CodePackage


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser(
        "verify",
        help="verify the trustworthiness of a running Cosmian Enclave web application "
        "and get the RA-TLS certificate",
    )

    parser.add_argument(
        "--evidence",
        required=True,
        type=Path,
        metavar="FILE",
        help="path to the evidence file",
    )

    parser.add_argument(
        "--package",
        type=Path,
        required=True,
        help="tarball file containing the container images and the code",
    )

    parser.add_argument(
        "--output",
        type=Path,
        required=True,
        help="output path of the verified RA-TLS certificate",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    if not args.output.is_dir():
        raise NotADirectoryError(f"{args.output} does not exist")

    workspace = Path(tempfile.mkdtemp(dir=args.output))
    log_path = workspace / "docker.log"

    evidence = ApplicationEvidence.load(args.evidence)

    LOG.info("Extracting the package at %s...", workspace)
    package = CodePackage.extract(workspace, args.package)

    LOG.info("A log file is generating at: %s", log_path)

    client = get_client_docker()
    image = load_docker_image(client, package.image_tar)
    mrenclave = compute_mr_enclave(
        client,
        image,
        evidence.input_args,
        workspace,
        log_path,
    )

    LOG.info("Fingerprint is: %s", mrenclave)

    try:
        verify_enclave(
            evidence.signer_pk,
            evidence.ratls_certificate,
            fingerprint=mrenclave,
            collaterals=evidence.collaterals,
        )
    except Exception as exc:
        LOG.error("Verification failed!")
        raise exc

    LOG.info("Verification successful")

    ratls_cert_path = args.output.resolve() / "ratls.pem"
    ratls_cert_path.write_bytes(
        evidence.ratls_certificate.public_bytes(encoding=Encoding.PEM)
    )

    LOG.info("The RA-TLS certificate has been saved at: %s", ratls_cert_path)

    # Clean up the workspace
    LOG.info("Cleaning up the temporary workspace...")
    shutil.rmtree(workspace)



"""cenclave.core.no_sgx_docker module."""

from pathlib import Path
from typing import ClassVar, Dict, List, Optional
from uuid import UUID

from pydantic import BaseModel

from cenclave.core.sgx_docker import SgxDockerConfig


class NoSgxDockerConfig(BaseModel):
    """Definition of container running on a non-sgx hardware."""

    subject: str
    subject_alternative_name: str
    expiration_date: Optional[int]
    client_certificate: Optional[str]
    size: int
    app_id: UUID
    application: str

    app_mountpoint: ClassVar[str] = "/opt/input"
    entrypoint: ClassVar[str] = "cenclave-run"

    def cmd(self) -> List[str]:
        """Serialize the docker command args."""
        args = [
            "--size",
            f"{self.size}M",
            "--subject",
            self.subject,
            "--san",
            self.subject_alternative_name,
            "--id",
            str(self.app_id),
            "--application",
            self.application,
            "--dry-run",
        ]

        if self.expiration_date:
            args.append("--expiration")
            args.append(str(self.expiration_date))

        if client_certificate := self.client_certificate:
            args.append("--client-certificate")
            args.append(client_certificate)

        return args

    def volumes(self, app_path: Path) -> Dict[str, Dict[str, str]]:
        """Define the docker volumes."""
        return {
            f"{app_path.resolve()}": {
                "bind": NoSgxDockerConfig.app_mountpoint,
                "mode": "rw",
            }
        }

    @staticmethod
    def from_sgx(docker_config: SgxDockerConfig):
        """Load from a SgxDockerConfig object."""
        return NoSgxDockerConfig(
            subject=docker_config.subject,
            subject_alternative_name=docker_config.subject_alternative_name,
            expiration_date=docker_config.expiration_date,
            client_certificate=docker_config.client_certificate,
            size=docker_config.size,
            app_id=docker_config.app_id,
            application=docker_config.application,
        )



"""Test model/package.py."""

import filecmp
from pathlib import Path
from tarfile import TarFile

import pytest

from cenclave.model.package import CodePackage


def test_create(workspace: Path):
    """Test `create` method."""
    package = CodePackage(
        code_tar=Path(__file__).parent / "data" / "package" / "app.tar",
        image_tar=Path(__file__).parent / "data" / "package" / "image.tar",
        test_tar=Path(__file__).parent / "data" / "package" / "tests.tar",
        config_path=Path(__file__).parent / "data" / "config.toml",
    )

    package_tar_ref = Path(__file__).parent / "data" / "package" / "package.tar"
    package_tar = workspace / "package.tar"
    package.create(package_tar)

    assert TarFile(package_tar).getnames() == TarFile(package_tar_ref).getnames()


def test_extract(workspace: Path):
    """Test the `extract` method."""
    package_tar = Path(__file__).parent / "data" / "package" / "package.tar"
    package = CodePackage.extract(workspace, package_tar)

    assert package.code_tar == workspace / "app.tar"
    assert package.image_tar == workspace / "image.tar"
    assert package.test_tar == workspace / "tests.tar"
    assert package.config_path == workspace / "config.toml"

    assert filecmp.cmp(
        Path(__file__).parent / "data" / "package" / "app.tar", package.code_tar
    )
    assert filecmp.cmp(
        Path(__file__).parent / "data" / "package" / "image.tar", package.image_tar
    )
    assert filecmp.cmp(
        Path(__file__).parent / "data" / "package" / "tests.tar", package.test_tar
    )
    assert filecmp.cmp(Path(__file__).parent / "data" / "config.toml", package.config_path)


def test_extract_bad_tar(workspace: Path):
    """Test the `extract` method with errors."""
    # Not a tar
    package_tar = Path(__file__).parent / "data" / "evidence.json"

    with pytest.raises(Exception):
        CodePackage.extract(workspace, package_tar)

    # Tar but without the expecting content
    package_tar = Path(__file__).parent / "data" / "package" / "app.tar"

    with pytest.raises(Exception):
        CodePackage.extract(workspace, package_tar)



"""cenclave.command.sgx_operator.stop module."""

from cenclave.command.helpers import get_app_container, get_client_docker
from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser(
        "stop", help="stop and optionally remove a running container"
    )

    parser.add_argument(
        "name",
        type=str,
        help="name of the application",
    )

    parser.add_argument(
        "--remove",
        action="store_true",
        help="remove the docker after being stopped, preventing it to be restarted later",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    client = get_client_docker()
    container = get_app_container(client, args.name)

    LOG.info("Stopping your application docker...")
    container.stop(timeout=1)

    LOG.info("Docker '%s' has been stopped!", args.name)

    if args.remove:
        container.remove()
        LOG.info("Docker '%s' has been removed!", args.name)



"""cenclave.command.sgx_operator.status module."""

from datetime import datetime

import requests

from cenclave.command.helpers import get_app_container, get_client_docker, is_running
from cenclave.core.sgx_docker import SgxDockerConfig
from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser("status", help="print the container status")

    parser.add_argument(
        "name",
        type=str,
        help="name of the application",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    client = get_client_docker()
    container = get_app_container(client, args.name)

    docker = SgxDockerConfig.load(container.attrs, container.labels)

    expires_at = datetime.fromtimestamp(docker.expiration_date)
    remaining_days = expires_at - datetime.now()

    LOG.info("    App name = %s", args.name)
    LOG.info("Enclave size = %dM", docker.size)
    LOG.info(" Common name = %s", docker.subject_alternative_name)
    LOG.info("        Port = %d", docker.port)
    LOG.info(" Healthcheck = %s", docker.healthcheck)
    LOG.info(
        "      Status = %s",
        (
            app_state(docker.host, docker.port, docker.healthcheck)
            if is_running(container)
            else container.status
        ),
    )
    LOG.info(
        "  Started at = %s",
        container.attrs["State"]["StartedAt"],
    )
    LOG.info(
        "  Expires at = %s (%d days remaining)",
        expires_at.astimezone(),
        remaining_days.days,
    )


def app_state(host: str, port: int, healthcheck_endpoint: str) -> str:
    """Determine the application state by querying it."""
    try:
        # Note: the configuration server allows any path
        # So: `healthcheck_endpoint`` does not exist but it's process as /
        # We can there do one query for the application and the configuration server
        response = requests.get(
            f"https://{host}:{port}{healthcheck_endpoint}",
            verify=False,
            timeout=60,
        )

        if response.status_code == 503:
            return "initializing"

        if response.status_code == 500:
            return "on error"

        if response.status_code == 200 and "Mse-Status" in response.headers:
            return "waiting secret keys"

        if response.status_code == 200:
            return "running"

        return "unknown"

    except requests.exceptions.SSLError:
        return "initializing"



"""Test model/evidence.py."""

import base64
import filecmp
from pathlib import Path

from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PublicFormat,
    load_pem_public_key,
)
from cryptography.x509 import load_pem_x509_certificate, load_pem_x509_crl

from cenclave.core.no_sgx_docker import NoSgxDockerConfig
from cenclave.model.evidence import ApplicationEvidence


def test_load():
    """Test `load` function."""
    json_path = Path(__file__).parent / "data/evidence.json"
    conf = ApplicationEvidence.load(path=json_path)

    ref_conf = ApplicationEvidence(
        input_args=NoSgxDockerConfig(
            host="localhost",
            subject="CN=localhost,O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
            subject_alternative_name="localhost",
            expiration_date=1714058115,
            size=4096,
            app_id="63322f85-1ff8-4483-91ae-f18d7398d157",
            application="app:app",
        ),
        ratls_certificate=load_pem_x509_certificate(
            b"-----BEGIN CERTIFICATE-----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\npLMzjW4ikTqPVuAq0bhKzA==\n-----END CERTIFICATE-----\n"
        ),
        root_ca_crl=load_pem_x509_crl(
            b"-----BEGIN X509 CRL-----\nMIIBITCByAIBATAKBggqhkjOPQQDAjBoMRowGAYDVQQDDBFJbnRlbCBTR1ggUm9v\ndCBDQTEaMBgGA1UECgwRSW50ZWwgQ29ycG9yYXRpb24xFDASBgNVBAcMC1NhbnRh\nIENsYXJhMQswCQYDVQQIDAJDQTELMAkGA1UEBhMCVVMXDTIzMDQwMzEwMjI1MVoX\nDTI0MDQwMjEwMjI1MVqgLzAtMAoGA1UdFAQDAgEBMB8GA1UdIwQYMBaAFCJlDNZa\nnTSJ84O0lVK/UBs5JwasMAoGCCqGSM49BAMCA0gAMEUCIFFXfUfZ+6FXtl8etfRl\ne7xeVsyvc1oD8blj1wSAWrEYAiEAk5AV7BY25+r6X0JsHkAmR8ZzEytoUMq9aM72\nutdoKgM=\n-----END X509 CRL-----\n"
        ),
        pck_platform_crl=load_pem_x509_crl(
            b"-----BEGIN X509 CRL-----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\nll0lzk4y1TQ9DtdJdB0IYrdG00OxLQ==\n-----END X509 CRL-----\n"
        ),
        tcb_info=base64.b64decode(
            " ".encode(
                "utf-8"
            )
        ),
        qe_identity=base64.b64decode(
            "eyJlbmNsYXZlSWRlbnRpdHkiOnsiaWQiOiJRRSIsInZlcnNpb24iOjIsImlzc3VlRGF0ZSI6IjIwMjMtMDYtMjBUMDA6MjQ6MDJaIiwibmV4dFVwZGF0ZSI6IjIwMjMtMDctMjBUMDA6MjQ6MDJaIiwidGNiRXZhbHVhdGlvbkRhdGFOdW1iZXIiOjE1LCJtaXNjc2VsZWN0IjoiMDAwMDAwMDAiLCJtaXNjc2VsZWN0TWFzayI6IkZGRkZGRkZGIiwiYXR0cmlidXRlcyI6IjExMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwMDAwIiwiYXR0cmlidXRlc01hc2siOiJGQkZGRkZGRkZGRkZGRkZGMDAwMDAwMDAwMDAwMDAwMCIsIm1yc2lnbmVyIjoiOEM0RjU3NzVENzk2NTAzRTk2MTM3Rjc3QzY4QTgyOUEwMDU2QUM4REVENzAxNDBCMDgxQjA5NDQ5MEM1N0JGRiIsImlzdnByb2RpZCI6MSwidGNiTGV2ZWxzIjpbeyJ0Y2IiOnsiaXN2c3ZuIjo4fSwidGNiRGF0ZSI6IjIwMjMtMDItMTVUMDA6MDA6MDBaIiwidGNiU3RhdHVzIjoiVXBUb0RhdGUifSx7InRjYiI6eyJpc3Zzdm4iOjZ9LCJ0Y2JEYXRlIjoiMjAyMS0xMS0xMFQwMDowMDowMFoiLCJ0Y2JTdGF0dXMiOiJPdXRPZkRhdGUiLCJhZHZpc29yeUlEcyI6WyJJTlRFTC1TQS0wMDYxNSJdfSx7InRjYiI6eyJpc3Zzdm4iOjV9LCJ0Y2JEYXRlIjoiMjAyMC0xMS0xMVQwMDowMDowMFoiLCJ0Y2JTdGF0dXMiOiJPdXRPZkRhdGUiLCJhZHZpc29yeUlEcyI6WyJJTlRFTC1TQS0wMDQ3NyIsIklOVEVMLVNBLTAwNjE1Il19LHsidGNiIjp7ImlzdnN2biI6NH0sInRjYkRhdGUiOiIyMDE5LTExLTEzVDAwOjAwOjAwWiIsInRjYlN0YXR1cyI6Ik91dE9mRGF0ZSIsImFkdmlzb3J5SURzIjpbIklOVEVMLVNBLTAwMzM0IiwiSU5URUwtU0EtMDA0NzciLCJJTlRFTC1TQS0wMDYxNSJdfSx7InRjYiI6eyJpc3Zzdm4iOjJ9LCJ0Y2JEYXRlIjoiMjAxOS0wNS0xNVQwMDowMDowMFoiLCJ0Y2JTdGF0dXMiOiJPdXRPZkRhdGUiLCJhZHZpc29yeUlEcyI6WyJJTlRFTC1TQS0wMDIxOSIsIklOVEVMLVNBLTAwMjkzIiwiSU5URUwtU0EtMDAzMzQiLCJJTlRFTC1TQS0wMDQ3NyIsIklOVEVMLVNBLTAwNjE1Il19LHsidGNiIjp7ImlzdnN2biI6MX0sInRjYkRhdGUiOiIyMDE4LTA4LTE1VDAwOjAwOjAwWiIsInRjYlN0YXR1cyI6Ik91dE9mRGF0ZSIsImFkdmlzb3J5SURzIjpbIklOVEVMLVNBLTAwMjAyIiwiSU5URUwtU0EtMDAyMTkiLCJJTlRFTC1TQS0wMDI5MyIsIklOVEVMLVNBLTAwMzM0IiwiSU5URUwtU0EtMDA0NzciLCJJTlRFTC1TQS0wMDYxNSJdfV19LCJzaWduYXR1cmUiOiI2NDMwZTgzNjNhNTNlNzc2MGM4MzllMDg2YmI3ZTliZjliZmFkNzQ0MmYzMjBkMDRjMDI0MDY2NTViYjQxNzc0ZWUwMjc4YWViY2Y0Mjk2YjUzZTQ2NDQ3OTc2NTZlYjNjNzVhYTRmNzE2YTAzODkyZmRjMGE3MzY3MTY5YmU2OCJ9".encode(
                "utf-8"
            )
        ),
        tcb_cert=load_pem_x509_certificate(
            b"-----BEGIN CERTIFICATE-----\nMIICizCCAjKgAwIBAgIUfjiC1ftVKUpASY5FhAPpFJG99FUwCgYIKoZIzj0EAwIw\naDEaMBgGA1UEAwwRSW50ZWwgU0dYIFJvb3QgQ0ExGjAYBgNVBAoMEUludGVsIENv\ncnBvcmF0aW9uMRQwEgYDVQQHDAtTYW50YSBDbGFyYTELMAkGA1UECAwCQ0ExCzAJ\nBgNVBAYTAlVTMB4XDTE4MDUyMTEwNTAxMFoXDTI1MDUyMTEwNTAxMFowbDEeMBwG\nA1UEAwwVSW50ZWwgU0dYIFRDQiBTaWduaW5nMRowGAYDVQQKDBFJbnRlbCBDb3Jw\nb3JhdGlvbjEUMBIGA1UEBwwLU2FudGEgQ2xhcmExCzAJBgNVBAgMAkNBMQswCQYD\nVQQGEwJVUzBZMBMGByqGSM49AgEGCCqGSM49AwEHA0IABENFG8xzydWRfK92bmGv\nP+mAh91PEyV7Jh6FGJd5ndE9aBH7R3E4A7ubrlh/zN3C4xvpoouGlirMba+W2lju\nypajgbUwgbIwHwYDVR0jBBgwFoAUImUM1lqdNInzg7SVUr9QGzknBqwwUgYDVR0f\nBEswSTBHoEWgQ4ZBaHR0cHM6Ly9jZXJ0aWZpY2F0ZXMudHJ1c3RlZHNlcnZpY2Vz\nLmludGVsLmNvbS9JbnRlbFNHWFJvb3RDQS5kZXIwHQYDVR0OBBYEFH44gtX7VSlK\nQEmORYQD6RSRvfRVMA4GA1UdDwEB/wQEAwIGwDAMBgNVHRMBAf8EAjAAMAoGCCqG\nSM49BAMCA0cAMEQCIB9C8wOAN/ImxDtGACV246KcqjagZOR0kyctyBrsGGJVAiAj\nftbrNGsGU8YH211dRiYNoPPu19Zp/ze8JmhujB0oBw==\n-----END CERTIFICATE-----\n"
        ),
        signer_pk=load_pem_public_key(
            b"-----BEGIN PUBLIC KEY-----\nMIIBoDANBgkqhkiG9w0BAQEFAAOCAY0AMIIBiAKCAYEAwJ1w0FTxi683hkxrg9fG\nSZ6AvDbfvsh52dPgrgRIB9jkIW1eJacXyuNZHznOKl+T/QaHorfNg+e6IB9Jaa80\nyeqp7lewtF7nG4CB83qSeU/GbehuYQeuTlIPvIZrrMl9DviiADOu7cCvRTVLcyj8\n1pKuWpWuUC0LVE8o6JsG8Zk89Qj653oJl+926koOAaQZ203w2fKhbXTEoyaaysbe\n2jzabuPWIFafxPiSO6p+UcXs+CzJfaZkVKBWAhHBZN5DYr+JhYqOJlb3bvsAxGuS\n4mWgxx1OGRXTJq1uHQvZOuOprUICqiYoc6UJAT5MiMOofwng498KXLl0Yg4qAYI3\nDCdYEiFC5URJzvlQzQzHV6q4mTep2YXtIUcCsiomFM3ldZVEXGtpST+qqQUKIEiC\n3nqLhX5YogqFbOm/xhE/RCgVWb7J4aSW9/+ohwwsRM7eRFTFylBv2BLcP5gl/f9P\nuMX6e9Bd/TxRj5tkk7BLfex6gdfVRWnJ1Odhht1AyyYXAgED\n-----END PUBLIC KEY-----\n"
        ),
    )

    assert conf.input_args == ref_conf.input_args
    assert conf.ratls_certificate == ref_conf.ratls_certificate
    assert conf.root_ca_crl == ref_conf.root_ca_crl
    assert conf.pck_platform_crl == ref_conf.pck_platform_crl
    assert conf.signer_pk.public_bytes(
        encoding=Encoding.PEM,
        format=PublicFormat.SubjectPublicKeyInfo,
    ) == ref_conf.signer_pk.public_bytes(
        encoding=Encoding.PEM,
        format=PublicFormat.SubjectPublicKeyInfo,
    )
    assert conf.tcb_info == ref_conf.tcb_info
    assert conf.qe_identity == ref_conf.qe_identity
    assert conf.tcb_cert == ref_conf.tcb_cert


def test_save(workspace: Path):
    """Test the `save` method."""
    json = Path(__file__).parent / "data/evidence.json"
    conf = ApplicationEvidence.load(path=json)

    tmp_json = workspace / "evidence.json"
    conf.save(tmp_json)

    assert filecmp.cmp(json, tmp_json)



"""cenclave.command.sgx_operator.stop module."""

from cenclave.command.helpers import get_app_container, get_client_docker
from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser("restart", help="restart a web application")

    parser.add_argument(
        "name",
        type=str,
        help="the name of the web application",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    client = get_client_docker()
    container = get_app_container(client, args.name)

    LOG.info("Restarting your application docker...")
    container.restart()
    LOG.info("Docker '%s' is now restarted!", args.name)



"""cenclave.common.helpers module."""

import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Optional

import importlib_resources
from docker.errors import NotFound
from docker.models.containers import Container
from jinja2 import Template

from cenclave.core.bootstrap import is_ready
from cenclave.core.clock_tick import ClockTick
from cenclave.core.conf import AppConf, AppConfParsingOption
from cenclave.core.test_docker import TestDockerConfig
from cenclave.error import AppContainerError
from cenclave.log import LOGGER as LOG
from cenclave.model.package import (
    DEFAULT_CODE_DIR,
    DEFAULT_CONFIG_FILENAME,
    DEFAULT_TEST_DIR,
)


def scaffold(app_name: str) -> Path:
    """Run the subcommand."""
    project_dir = Path(os.getcwd()) / app_name

    # Copy the template files
    shutil.copytree(importlib_resources.files("cenclave") / "template", project_dir)

    template_conf_file = project_dir / "config.toml.template"
    conf_file = project_dir / DEFAULT_CONFIG_FILENAME

    # Initialize the configuration file
    tm = Template(template_conf_file.read_text())
    content = tm.render(name=app_name)

    conf_file.write_text(content)
    template_conf_file.unlink()

    app_conf = AppConf.load(conf_file, option=AppConfParsingOption.SkipCloud)

    # Initialize the python code file
    code_dir = project_dir / DEFAULT_CODE_DIR
    template_code_file = code_dir / (app_conf.python_module + ".py.template")
    code_file = template_code_file.with_suffix("")

    tm = Template(template_code_file.read_text())
    content = tm.render(
        app=app_conf.python_variable,
        healthcheck_endpoint=app_conf.healthcheck_endpoint,
    )
    code_file.write_text(content)
    template_code_file.unlink()

    # Initialize the .enclaveignore
    ignore_file: Path = code_dir / "dotcenclaveignore"
    ignore_file.rename(code_dir / ".cenclaveignore")

    # Initialize the pytest code files
    pytest_dir = project_dir / DEFAULT_TEST_DIR
    template_pytest_file = pytest_dir / "conftest.py.template"
    pytest_file = template_pytest_file.with_suffix("")

    tm = Template(template_pytest_file.read_text())
    content = tm.render()
    pytest_file.write_text(content)
    template_pytest_file.unlink()

    pytest_dir = project_dir / DEFAULT_TEST_DIR
    template_pytest_file = pytest_dir / "test_app.py.template"
    pytest_file = template_pytest_file.with_suffix("")

    tm = Template(template_pytest_file.read_text())
    content = tm.render(
        healthcheck_endpoint=app_conf.healthcheck_endpoint,
    )
    pytest_file.write_text(content)
    template_pytest_file.unlink()

    return conf_file


# pylint: disable=too-many-positional-arguments
def try_run_test_docker(
    app_config: AppConf,
    client,
    docker_name,
    container_name: str,
    docker_config: TestDockerConfig,
    test_path: Path,
    secrets_path: Optional[Path],
    sealed_secrets_path: Optional[Path],
    ignore_tests: bool = False,
):
    """Try to start the app docker to test."""
    success = False
    try:
        container = run_app_test_docker(
            client,
            docker_name,
            container_name,
            docker_config,
            app_config.healthcheck_endpoint,
        )

        if not ignore_tests:
            success = run_tests(
                app_config,
                test_path,
                secrets_path,
                sealed_secrets_path,
            )
        else:
            LOG.info(
                "The docker '%s' is started. "
                "You can now query your application on http://localhost:%s",
                docker_name,
                docker_config.port,
            )

            success = True

            for line in container.logs(stream=True):
                LOG.info(line.decode("utf-8").strip())

    except Exception as exc:
        raise exc
    finally:
        try:
            container = client.containers.get(container_name)
            if not success:
                LOG.info("The docker logs are:\n%s", container.logs().decode("utf-8"))
            container.stop(timeout=1)
            # We need to remove the container since we declare remove=False previously
            container.remove()
        except NotFound:
            pass


def run_app_test_docker(
    client,
    docker_name,
    container_name: str,
    docker_config: TestDockerConfig,
    healthcheck_endpoint: str,
) -> Container:
    """Run the app docker to test."""
    container = client.containers.run(
        docker_name,
        name=container_name,
        command=docker_config.cmd(),
        volumes=docker_config.volumes(),
        entrypoint=TestDockerConfig.entrypoint,
        ports=docker_config.ports(),
        detach=True,
        # We do not remove the container to be able to print the error (if some)
        remove=False,
    )

    clock = ClockTick(
        period=5,
        timeout=60,
        message="Test application docker is unreachable!",
    )

    while clock.tick():
        # Note: container.status is not actualized.
        # Get it again to have the current status
        container = client.containers.get(container_name)

        if container.status == "exited":
            raise AppContainerError("Application docker fails to start")

        if is_ready(f"http://localhost:{docker_config.port}", healthcheck_endpoint):
            break

    return container


def run_tests(
    app_config: AppConf,
    tests: Path,
    secrets: Optional[Path],
    sealed_secrets: Optional[Path],
) -> bool:
    """Run the tests."""
    LOG.info("Installing tests requirements...")
    for package in app_config.tests_requirements:
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", package],
            stdout=subprocess.DEVNULL,
        )

    LOG.info("Running tests...")
    env = dict(os.environ)
    if secrets:
        env["TEST_SECRET_JSON"] = str(secrets.resolve())

    if sealed_secrets:
        env["TEST_SEALED_SECRET_JSON"] = str(sealed_secrets.resolve())

    try:
        subprocess.check_call(
            [sys.executable, "-m", app_config.tests_cmd], cwd=tests, env=env
        )

        LOG.info("Tests successful")
        return True
    except subprocess.CalledProcessError:
        LOG.error("Tests failed!")

    return False



"""cenclave.command.code_provider.keygen module."""

import base64
from pathlib import Path

from cenclave_lib_crypto.x25519 import x25519_keygen
from cenclave_lib_crypto.xsalsa20_poly1305 import random_key

from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser(
        "keygen", help="key generation for symmetric or asymmetric cryptography"
    )

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "--symmetric",
        action="store_true",
        help="generate a symmetric key for XSalsa20-Poly1305",
    )

    group.add_argument(
        "--asymmetric",
        action="store_true",
        help="generate an asymmetric keypair on Curve25519",
    )

    parser.add_argument(
        "--output",
        type=Path,
        metavar="FILE",
        help="path to write the generated key(s)",
    )

    parser.add_argument(
        "--force",
        action="store_true",
        help="overwrite if output file exists",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    output_path: Path = Path(args.output).resolve()

    if output_path.exists() and not args.force:
        raise FileExistsError(f"{output_path} already exists")

    if args.symmetric:
        output_path.write_bytes(random_key())
        LOG.info("Symmetric key wrote to %s", output_path)
    else:  # args.asymmetric
        pk, sk = x25519_keygen()
        LOG.info("Public key: %s", pk.hex())
        LOG.info("Public key (Base64): %s", base64.b64encode(pk).decode("utf-8"))

        output_path.write_bytes(pk + sk)
        LOG.info("Keypair wrote to %s", output_path)



"""cenclave.command.code_provider.scaffold module."""

from cenclave.common.helpers import scaffold
from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser(
        "scaffold", help="create a new boilerplate web application"
    )

    parser.add_argument(
        "app_name",
        type=str,
        help="name of the Python web application",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    conf_file = scaffold(args.app_name)

    LOG.info(  # type: ignore
        "An example app has been generated in the directory: %s/", args.app_name
    )
    LOG.warning("You can configure your web application in: %s", conf_file)



"""cenclave.core.spinner module."""

import itertools
import sys
import threading
import time


class Spinner:
    """Class to wait and print a spinner."""

    def __init__(self, message: str):
        """Initialize the spinner."""
        self.spinner = itertools.cycle(["-", "/", "|", "\\"])
        self.busy = True
        self.period = 1
        self.message = message

    def __task(self):
        """Make the spinner spin."""
        while self.busy:
            sys.stdout.write(next(self.spinner))  # write the next character
            sys.stdout.flush()  # flush stdout buffer (actual character display)
            sys.stdout.write("\b")  # erase the last written char
            time.sleep(self.period)  # wait for a period time

    def start(self):
        """Start spinning."""
        self.busy = True
        sys.stdout.write(self.message)
        threading.Thread(target=self.__task).start()

    def __enter__(self):
        """Entrypoint of the `with` statement."""
        self.start()
        return self

    def stop(self):
        """Remove the spinner."""
        self.busy = False
        sys.stdout.write(" ")
        sys.stdout.write("\n")
        sys.stdout.flush()  # flush stdout buffer (actual character display)

    def __exit__(self, exception_type, exception_value, traceback):
        """Endpoint of the `with` statement."""
        self.stop()



"""Test base64.py."""

from cenclave.core.base64 import base64url_decode, base64url_encode


def test_encode():
    """Test `base64url_encode`."""
    assert base64url_encode(b"") == ""
    assert base64url_encode(b"a") == "YQ"
    assert base64url_encode(b"ab") == "YWI"
    assert base64url_encode(b"abc") == "YWJj"
    assert base64url_encode(b"abcd") == "YWJjZA"
    assert base64url_encode(b"abcde") == "YWJjZGU"


def test_decode():
    """Test `base64url_decode`."""
    assert base64url_decode("") == b""
    assert base64url_decode("YQ") == b"a"
    assert base64url_decode("YWI") == b"ab"
    assert base64url_decode("YWJj") == b"abc"
    assert base64url_decode("YWJjZA") == b"abcd"
    assert base64url_decode("YWJjZGU") == b"abcde"



"""cenclave.command.code_provider.package module."""

import argparse
import os
import shutil
import tempfile
import time
from pathlib import Path
from typing import Dict, Optional, Tuple

from cenclave_lib_crypto.xsalsa20_poly1305 import encrypt_directory, random_key
from docker.errors import BuildError

from cenclave.command.helpers import get_client_docker
from cenclave.core.conf import AppConf, AppConfParsingOption
from cenclave.core.fs import tar, whitelist
from cenclave.core.ignore_file import IgnoreFile
from cenclave.log import LOGGER as LOG
from cenclave.model.package import (
    CODE_TAR_NAME,
    DEFAULT_CODE_DIR,
    DEFAULT_CONFIG_FILENAME,
    DEFAULT_DOCKERFILE_FILENAME,
    DEFAULT_TEST_DIR,
    DOCKER_IMAGE_TAR_NAME,
    TEST_TAR_NAME,
    CodePackage,
)


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser(
        "package",
        help="generate a package containing the container image and the code to "
        "run with Cosmian Enclave",
    )

    parser.add_argument("--project", type=Path, help="path of the project to pack")

    parser.add_argument("--code", type=Path, help="path of the code to run")

    parser.add_argument("--config", type=Path, help="path to the code configuration")

    parser.add_argument("--dockerfile", type=Path, help="path to the Dockerfile")

    parser.add_argument("--test", type=Path, help="test directory")

    parser.add_argument(
        "--output", type=Path, required=True, help="directory to write the package"
    )

    parser.add_argument(
        "--encrypt",
        action="store_true",
        help="encrypt the code before packaging it",
    )

    parser.set_defaults(func=run)


# pylint: disable=too-many-statements
def run(args) -> None:
    """Run the subcommand."""
    package_path: Path = args.output.resolve()
    if not package_path.exists():
        os.makedirs(package_path)
    elif not package_path.is_dir():
        raise NotADirectoryError(f"{package_path} does not exist")

    code_path: Path
    test_path: Path
    config_path: Path
    dockerfile_path: Path

    if args.project:
        if any([args.code, args.config, args.dockerfile, args.test]):
            raise argparse.ArgumentTypeError(
                "[--project] and [--code & --config & --dockerfile & --test] "
                "are mutually exclusive"
            )

        if not args.project.is_dir():
            raise NotADirectoryError(f"`{args.project}` does not exist")

        code_path = args.project / DEFAULT_CODE_DIR
        test_path = args.project / DEFAULT_TEST_DIR
        config_path = args.project / DEFAULT_CONFIG_FILENAME
        dockerfile_path = args.project / DEFAULT_DOCKERFILE_FILENAME
    else:
        if not all([args.code, args.config, args.dockerfile, args.test]):
            raise argparse.ArgumentTypeError(
                "the following arguments are required: "
                "--code, --config, --dockerfile, --test"
            )

        code_path = args.code
        test_path = args.test
        config_path = args.config
        dockerfile_path = args.dockerfile

    if not code_path.is_dir():
        raise NotADirectoryError(f"`{code_path}` does not exist")

    if not test_path.is_dir():
        raise NotADirectoryError(f"`{test_path}` does not exist")

    if not config_path.is_file():
        raise FileNotFoundError(f"`{config_path}` does not exist")

    if not dockerfile_path.is_file():
        raise FileNotFoundError(f"`{dockerfile_path}` does not exist")

    code_config = AppConf.load(config_path, option=AppConfParsingOption.SkipCloud)

    workspace = Path(tempfile.mkdtemp(dir=package_path))

    LOG.info("A workspace has been created at: %s", str(workspace))

    package = CodePackage(
        code_tar=workspace / CODE_TAR_NAME,
        image_tar=workspace / DOCKER_IMAGE_TAR_NAME,
        test_tar=workspace / TEST_TAR_NAME,
        config_path=config_path.resolve(),
    )

    now = time.time_ns()
    code_secret_path = package_path / f"package_{code_config.name}_{now}.key"
    package_path = package_path / f"package_{code_config.name}_{now}.tar"

    (secret_key, _) = create_code_tar(
        code_path.resolve(), package.code_tar, args.encrypt
    )

    if secret_key:
        code_secret_path.write_bytes(secret_key)
        LOG.info("Your code secret key has been saved at: %s", code_secret_path)

    create_test_tar(
        test_path.resolve(),
        package.test_tar,
    )

    create_image_tar(dockerfile_path.resolve(), code_config.name, package.image_tar)

    LOG.info("Creating the final package...")

    package.create(package_path)

    LOG.info("Your package is now ready to be shared: %s", package_path)

    # Clean up the workspace
    LOG.info("Cleaning up the temporary workspace...")
    shutil.rmtree(workspace)


def create_code_tar(
    code_path: Path, output_tar_path: Path, encrypt_code: bool
) -> Tuple[Optional[bytes], Optional[Dict[str, bytes]]]:
    """Create the tarball for the code directory."""
    if encrypt_code:
        LOG.info("Encrypting your code...")

        # Generate the key to encrypt the code
        secret_key = random_key()

        encrypted_path = output_tar_path.parent / "encrypted_code"

        # Encrypt the code directory
        nounces = encrypt_directory(
            dir_path=code_path,
            pattern="*",
            key=secret_key,
            nonces=None,
            exceptions=whitelist(),
            ignore_patterns=list(IgnoreFile.parse(code_path)),
            out_dir_path=encrypted_path,
        )

        LOG.info("Your encryption key is: %s", bytes(secret_key).hex())
        LOG.info("Building the code archive...")

        # Generate the tarball
        tar(dir_path=encrypted_path, tar_path=output_tar_path)

        return (secret_key, nounces)

    LOG.info("Building the code archive...")

    mirror_path = output_tar_path.parent / "mirrored_code"

    # We copy the code directory to remove the files to ignore when taring
    shutil.copytree(
        code_path,
        mirror_path,
        ignore=shutil.ignore_patterns(*list(IgnoreFile.parse(code_path))),
    )

    # Generate the tarball
    tar(dir_path=mirror_path, tar_path=output_tar_path)

    return (None, None)


def create_test_tar(test_path: Path, output_tar_path: Path):
    """Create the tarball for the tests directory."""
    LOG.info("Building the tests archive...")

    mirror_path = output_tar_path.parent / "mirrored_tests"

    # We copy the code directory to remove the files to ignore when taring
    shutil.copytree(
        test_path,
        mirror_path,
        ignore=shutil.ignore_patterns(*["__pycache__", ".pytest_cache"]),
    )

    # Generate the tarball
    tar(dir_path=mirror_path, tar_path=output_tar_path)


def create_image_tar(dockerfile: Path, image_name: str, output_tar_path: Path):
    """Build the docker image and export it into a tarball."""
    client = get_client_docker()

    try:
        LOG.info("Building your docker image...")

        # Build the docker
        (image, _streamer) = client.images.build(
            path=str(dockerfile.parent),
            tag=f"{image_name}:{time.time_ns()}",
        )

        # for chunk in streamer:
        #     if "stream" in chunk:
        #         for line in chunk["stream"].splitlines():
        #             LOG.info(line)

        LOG.info("Building the image archive...")

        # Save it as a tarball
        with open(output_tar_path, "wb") as f:
            for chunk in image.save(named=True):
                f.write(chunk)

    except BuildError as exc:
        LOG.error("Failed to build your docker!")
        raise exc



"""cenclave.log module."""

import logging
import sys

LOGGER = logging.getLogger("cenclave")

LOGGING_SUCCESS = 25
LOGGING_ADVICE = 26


def setup_logging(debug: bool = False):
    """Configure basic logging."""
    format_msg = "%(message)s"

    # Define a specific format for stdout handler
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.DEBUG if debug else logging.INFO)

    logging.basicConfig(format=format_msg, handlers=[stdout_handler])
    LOGGER.setLevel(logging.DEBUG if debug else logging.INFO)

    # Add a success level to the default logger (then we can write LOG.success("msg"))
    logging.addLevelName(LOGGING_SUCCESS, "SUCCESS")
    # pylint: disable=protected-access
    setattr(
        LOGGER,
        "success",
        lambda message, *args: LOGGER._log(LOGGING_SUCCESS, message, args),
    )

    # Add an advice level to the default logger (then we can write LOG.advice("msg"))
    logging.addLevelName(LOGGING_ADVICE, "ADVICE")
    # pylint: disable=protected-access
    setattr(
        LOGGER,
        "advice",
        lambda message, *args: LOGGER._log(LOGGING_ADVICE, message, args),
    )



"""cenclave module."""

import os

__version__ = "1.0.0a9"

# PCCS to retrieve collaterals
PCCS_URL = os.getenv("PCCS_URL", default="https://pccs.cosmian.com")

# URL of Cosmian Enclave's documentation
DOC_URL = "https://docs.cosmian.com/compute/cosmian_enclave/overview/"



"""cenclave.command.code_provider.encrypt module."""

import sys
from pathlib import Path

from cryptography.fernet import Fernet

from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser(
        "encrypt", help="encrypt a file using Fernet symmetric encryption"
    )

    parser.add_argument(
        "--input",
        type=Path,
        required=True,
        help="path to the file to encrypt",
    )

    parser.add_argument(
        "--key",
        type=Path,
        metavar="FILE",
        required=True,
        help="path to the file within a 32 bytes key URL Safe Base64 encoded",
    )

    parser.add_argument(
        "--output",
        type=Path,
        metavar="FILE",
        help="path to write encrypted file",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    LOG.info("Encrypting %s...", args.input)

    key: bytes = args.key.read_bytes()
    data: bytes = args.input.read_bytes()

    encrypted_data: bytes = Fernet(key).encrypt(data)

    if args.output:
        args.output.write_bytes(encrypted_data)
        LOG.info("File encrypted to %s", args.output)
    else:
        LOG.info("Data sucessfully encrypted!")
        LOG.info(
            "----------------------------------------------------------------------"
        )
        sys.stdout.buffer.write(encrypted_data)



"""cenclave.command.sgx_operator.logs module."""

from cenclave.command.helpers import get_app_container, get_client_docker
from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser("logs", help="print container logs")

    parser.add_argument(
        "name",
        type=str,
        help="the name of the container",
    )

    parser.add_argument(
        "-f",
        "--follow",
        action="store_true",
        help="follow log output",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    client = get_client_docker()
    container = get_app_container(client, args.name)

    if args.follow:
        LOG.info("skipping...")
        for line in container.logs(tail=10, stream=True):
            LOG.info(line.decode("utf-8").strip())
    else:
        LOG.info(container.logs().decode("utf-8"))



"""cenclave.command.sgx_operator.run module."""

import json
from pathlib import Path

from cenclave.command.helpers import get_client_docker, get_running_app_container
from cenclave.core.bootstrap import (
    ConfigurationPayload,
    configure_app,
    is_waiting_for_secrets,
    wait_for_app_server,
)
from cenclave.core.clock_tick import ClockTick
from cenclave.core.sgx_docker import SgxDockerConfig
from cenclave.core.spinner import Spinner
from cenclave.error import AppContainerBadState
from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser(
        "run",
        help="finalise the configuration of the application "
        "docker and run the application code",
    )

    parser.add_argument(
        "name",
        type=str,
        help="name of the application",
    )

    parser.add_argument(
        "--secrets",
        type=Path,
        help="secrets.json file path",
    )

    parser.add_argument(
        "--sealed-secrets",
        type=Path,
        help="sealed secrets.json file path",
    )

    parser.add_argument(
        "--key",
        type=Path,
        help="code decryption sealed key file path",
    )

    parser.add_argument(
        "--timeout",
        type=int,
        required=False,
        default=24 * 60,
        help="stop the deployment if the application does not "
        "respond after a delay (in min). (Default: 1440 min)",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    client = get_client_docker()
    container = get_running_app_container(client, args.name)

    docker = SgxDockerConfig.load(container.attrs, container.labels)

    if not is_waiting_for_secrets(f"https://{docker.host}:{docker.port}", False):
        raise AppContainerBadState(
            "Your application is not waiting for secrets. Have you already set it?"
        )

    data = ConfigurationPayload(
        app_id=docker.app_id,
        secrets=json.loads(args.secrets.read_text()) if args.secrets else None,
        sealed_secrets=(
            args.sealed_secrets.read_bytes() if args.sealed_secrets else None
        ),
        code_secret_key=args.key.read_bytes() if args.key else None,
        ssl_private_key=None,
    )

    LOG.info("Sending data to the configuration server...")
    configure_app(
        f"https://{docker.host}:{docker.port}",
        data.payload(),
        False,
    )
    LOG.info("Your application is now configured!")

    with Spinner("Waiting for your application to be ready... "):
        wait_for_app_server(
            ClockTick(
                period=5,
                timeout=60 * args.timeout,
                message="Your application is unreachable!",
            ),
            f"https://{docker.host}:{docker.port}",
            docker.healthcheck,
            False,
            get_running_app_container,
            (
                client,
                args.name,
            ),
        )

    LOG.info("Application ready!")
    LOG.info("Feel free to test it using the `cenclave test` command")



"""cenclave.command.sgx_operator.list module."""

from cenclave.command.helpers import get_client_docker
from cenclave.core.sgx_docker import SgxDockerConfig
from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser("list", help="list the running containers")

    parser.set_defaults(func=run)


def run(_args) -> None:
    """Run the subcommand."""
    client = get_client_docker()

    containers = client.containers.list(
        all=True, filters={"label": SgxDockerConfig.docker_label}
    )

    LOG.info(
        "\n %s | %s | %s [Image name] ",
        "Started at".center(29),
        "Status".center(10),
        "Application name",
    )
    LOG.info(("-" * 86))

    for container in containers:
        LOG.info(
            "%30s | %s | %s [%s]",
            container.attrs["State"]["StartedAt"],
            container.status.center(10),
            container.name,
            container.image.tags[0],
        )



"""cenclave.command.code_provider.localtest module."""

import argparse
from pathlib import Path
from typing import Optional, cast

from docker.errors import BuildError

from cenclave.command.helpers import get_client_docker
from cenclave.common.helpers import try_run_test_docker
from cenclave.core.conf import AppConf, AppConfParsingOption
from cenclave.core.test_docker import TestDockerConfig
from cenclave.error import DockerBuildError
from cenclave.log import LOGGER as LOG
from cenclave.model.package import (
    DEFAULT_CODE_DIR,
    DEFAULT_CONFIG_FILENAME,
    DEFAULT_DOCKERFILE_FILENAME,
    DEFAULT_SEAL_SECRETS_FILENAME,
    DEFAULT_SECRETS_FILENAME,
    DEFAULT_TEST_DIR,
)


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser(
        "localtest", help="test locally a web app without SGX"
    )

    parser.add_argument("--project", type=Path, help="path of the project to test")

    parser.add_argument("--code", type=Path, help="path to the code to run")

    parser.add_argument(
        "--config", type=Path, help="conf path extracted from the tarball"
    )

    parser.add_argument("--dockerfile", type=Path, help="path to the Dockerfile")

    parser.add_argument(
        "--test",
        type=Path,
        help="path of the test directory extracted from the tarball package",
    )

    parser.add_argument(
        "--secrets",
        type=Path,
        help="the `secrets.json` file path",
    )

    parser.add_argument(
        "--sealed-secrets",
        type=Path,
        help="the secrets JSON to seal file path (unsealed for the test purpose)",
    )

    parser.add_argument(
        "--simu-enclave-keypair",
        type=Path,
        help="path with the keypair of the enclave for simulation",
    )

    parser.add_argument(
        "--no-tests",
        action="store_true",
        help="do not run the tests: just start the docker",
    )

    parser.set_defaults(func=run)


# pylint: disable=too-many-statements
def run(args) -> None:
    """Run the subcommand."""
    code_path: Path
    test_path: Path
    config_path: Path
    dockerfile_path: Path
    secrets_path: Optional[Path]
    sealed_secrets_path: Optional[Path]

    if args.project:
        if any([args.code, args.config, args.dockerfile, args.test]):
            raise argparse.ArgumentTypeError(
                "[--project] and [--code & --config & --dockerfile & --test] "
                "are mutually exclusive"
            )

        if not args.project.is_dir():
            raise NotADirectoryError(f"`{args.project}` does not exist")

        code_path = args.project / DEFAULT_CODE_DIR
        test_path = args.project / DEFAULT_TEST_DIR
        config_path = args.project / DEFAULT_CONFIG_FILENAME
        dockerfile_path = args.project / DEFAULT_DOCKERFILE_FILENAME

        secrets_path = args.project / DEFAULT_SECRETS_FILENAME
        sealed_secrets_path = args.project / DEFAULT_SEAL_SECRETS_FILENAME

    else:
        if not all([args.code, args.config, args.dockerfile, args.test]):
            raise argparse.ArgumentTypeError(
                "the following arguments are required: "
                "--code, --dockerfile, --test, --config\n"
                "the following arguments remain optional: "
                "[--secrets], [--sealed-secrets]"
            )

        code_path = args.code
        test_path = args.test
        config_path = args.config
        dockerfile_path = args.dockerfile
        secrets_path = args.secrets
        sealed_secrets_path = args.sealed_secrets

    if not code_path.is_dir():
        raise NotADirectoryError(f"`{code_path}` does not exist")

    if not test_path.is_dir():
        raise NotADirectoryError(f"`{test_path}` does not exist")

    if not config_path.is_file():
        raise FileNotFoundError(f"`{config_path}` does not exist")

    if not dockerfile_path.is_file():
        raise FileNotFoundError(f"`{dockerfile_path}` does not exist")

    if secrets_path and not secrets_path.is_file():
        raise FileNotFoundError(f"`{secrets_path}` does not exist")

    if sealed_secrets_path and not sealed_secrets_path.is_file():
        raise FileNotFoundError(f"`{sealed_secrets_path}` does not exist")

    enclave_sk_path: Optional[Path] = None

    if enclave_keypair_path := args.simu_enclave_keypair:
        enclave_sk: bytes = enclave_keypair_path.read_bytes()[32:]
        enclave_sk_path = enclave_keypair_path.parent / "enclave.key"
        cast(Path, enclave_sk_path).write_bytes(enclave_sk)

    code_config = AppConf.load(config_path, option=AppConfParsingOption.SkipCloud)
    container_name = docker_name = f"{code_config.name}_test"

    client = get_client_docker()

    build_test_docker(client, dockerfile_path, docker_name)

    LOG.info("Starting the docker: %s...", docker_name)
    docker_config = TestDockerConfig(
        code=code_path,
        application=code_config.python_application,
        secrets=secrets_path,
        sealed_secrets=sealed_secrets_path,
        simu_enclave_sk=enclave_sk_path,
        port=5000,
    )

    try_run_test_docker(
        code_config,
        client,
        docker_name,
        container_name,
        docker_config,
        test_path,
        secrets_path,
        sealed_secrets_path,
        args.no_tests,
    )


def build_test_docker(client, dockerfile: Path, docker_name: str):
    """Build the test docker."""
    try:
        LOG.info("Building your docker image...")

        # Build the docker
        client.images.build(
            path=str(dockerfile.parent),
            tag=docker_name,
        )
    except BuildError as exc:
        raise DockerBuildError(f"Failed to build your docker: {exc}") from exc



"""Test command/*.py."""

import base64
import io
import os
import re
import time
from argparse import Namespace
from pathlib import Path

import pytest
import requests
from conftest import capture_logs
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PublicFormat,
    PrivateFormat,
)

from cenclave.command.code_provider.encrypt import run as do_encrypt
from cenclave.command.code_provider.decrypt import run as do_decrypt
from cenclave.command.code_provider.localtest import run as do_test_dev
from cenclave.command.code_provider.package import run as do_package
from cenclave.command.code_provider.scaffold import run as do_scaffold
from cenclave.command.code_provider.seal import run as do_seal
from cenclave.command.code_provider.unseal import run as do_unseal
from cenclave.command.code_provider.verify import run as do_verify
from cenclave.command.sgx_operator.evidence import run as do_evidence
from cenclave.command.sgx_operator.list_all import run as do_list
from cenclave.command.sgx_operator.logs import run as do_logs
from cenclave.command.sgx_operator.restart import run as do_restart
from cenclave.command.sgx_operator.run import run as do_run
from cenclave.command.sgx_operator.spawn import run as do_spawn
from cenclave.command.sgx_operator.status import run as do_status
from cenclave.command.sgx_operator.stop import run as do_stop
from cenclave.command.sgx_operator.test import run as do_test


@pytest.mark.sgx
@pytest.mark.incremental
def test_scaffold(workspace, app_name):
    """Test the `scaffold` subcommand."""
    # Go into a unique tmp directory
    os.chdir(workspace)

    # Run scaffold
    do_scaffold(Namespace(**{"app_name": app_name}))

    # Check creation of files
    path = workspace / app_name
    pytest.app_path = path

    assert path.exists()
    assert (path / "Dockerfile").exists()
    assert (path / "config.toml").exists()
    assert (path / "secrets.json").exists()
    assert (path / "secrets_to_seal.json").exists()
    assert (path / "src").exists()
    assert (path / "src" / "app.py").exists()
    assert (path / "src" / ".cenclaveignore").exists()
    assert (path / "tests").exists()


@pytest.mark.sgx
@pytest.mark.incremental
def test_test_dev(cmd_log: io.StringIO):
    """Test the `test-dev` subcommand."""
    do_test_dev(
        Namespace(
            **{
                "project": None,
                "code": pytest.app_path / "src",
                "dockerfile": pytest.app_path / "Dockerfile",
                "config": pytest.app_path / "config.toml",
                "secrets": pytest.app_path / "secrets.json",
                "sealed_secrets": pytest.app_path / "secrets_to_seal.json",
                "test": pytest.app_path / "tests",
                "no_tests": False,
            }
        )
    )

    # Check the tar generation
    assert "Tests successful" in capture_logs(cmd_log)


@pytest.mark.sgx
@pytest.mark.incremental
def test_test_dev_project(cmd_log: io.StringIO):
    """Test the `test-dev` subcommand by specifying a project directory."""
    do_test_dev(
        Namespace(
            **{
                "project": pytest.app_path,
                "code": None,
                "dockerfile": None,
                "config": None,
                "secrets": None,
                "sealed_secrets": None,
                "test": None,
                "no_tests": False,
            }
        )
    )

    # Check the tar generation
    assert "Tests successful" in capture_logs(cmd_log)


@pytest.mark.sgx
@pytest.mark.incremental
def test_package(workspace: Path, cmd_log: io.StringIO):
    """Test the `package` subcommand."""
    do_package(
        Namespace(
            **{
                "project": None,
                "code": pytest.app_path / "src",
                "config": pytest.app_path / "config.toml",
                "dockerfile": pytest.app_path / "Dockerfile",
                "test": pytest.app_path / "tests",
                "encrypt": True,
                "output": workspace,
            }
        )
    )

    # Check the tar generation
    output = capture_logs(cmd_log)
    try:
        pytest.key_path = Path(
            re.search(
                "Your code secret key has been saved at: ([A-Za-z0-9/._-]+)", output
            ).group(1)
        )

        pytest.package_path = Path(
            re.search(
                "Your package is now ready to be shared: ([A-Za-z0-9/._-]+)", output
            ).group(1)
        )
    except AttributeError:
        print(output)
        assert False

    assert pytest.package_path.exists()


@pytest.mark.sgx
@pytest.mark.incremental
def test_package_project(workspace: Path, cmd_log: io.StringIO):
    """Test the `package` subcommand by specifying a project directory."""
    do_package(
        Namespace(
            **{
                "project": pytest.app_path,
                "code": None,
                "config": None,
                "dockerfile": None,
                "test": None,
                "encrypt": True,
                "output": workspace,
            }
        )
    )

    # Check the tar generation
    output = capture_logs(cmd_log)
    try:
        pytest.key_path = Path(
            re.search(
                "Your code secret key has been saved at: ([A-Za-z0-9/._-]+)", output
            ).group(1)
        )

        pytest.package_path = Path(
            re.search(
                "Your package is now ready to be shared: ([A-Za-z0-9/._-]+)", output
            ).group(1)
        )
    except AttributeError:
        print(output)
        assert False

    assert pytest.package_path.exists()


@pytest.mark.sgx
@pytest.mark.incremental
def test_spawn(
    workspace: Path,
    cmd_log: io.StringIO,
    app_name: str,
    port: int,
    host: str,
    signer_key: Path,
    pccs_url: str,
):
    """Test the `spawn` subcommand."""
    do_spawn(
        Namespace(
            **{
                "name": app_name,
                "pccs": pccs_url,
                "package": pytest.package_path,
                "host": host,
                "subject": f"CN={host},O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
                "san": host,
                "days": 2,
                "port": port,
                "size": 4096,
                "timeout": 5,
                "signer_key": signer_key,
                "output": workspace,
            }
        )
    )

    output = capture_logs(cmd_log)
    try:
        pytest.evidence_path = Path(
            re.search(
                "The evidence file has been generated at: ([A-Za-z0-9/._-]+)", output
            ).group(1)
        )
    except AttributeError:
        print(output)
        assert False

    assert pytest.evidence_path.exists()


@pytest.mark.sgx
@pytest.mark.incremental
def test_logs(cmd_log: io.StringIO, app_name: str):
    """Test the `logs` subcommand."""
    do_logs(Namespace(**{"name": app_name, "follow": False}))

    output = capture_logs(cmd_log)
    try:
        pytest.fingerprint = re.search(
            "Measurement:\n[ ]*([a-z0-9]{64})", output
        ).group(1)
    except AttributeError:
        print(output)
        assert False

    assert "Starting the configuration server..." in output


@pytest.mark.sgx
@pytest.mark.incremental
def test_status_conf_server(cmd_log: io.StringIO, app_name: str, port: int, host: str):
    """Test the `status` subcommand on the conf server."""
    do_status(
        Namespace(
            **{
                "name": app_name,
            }
        )
    )

    output = capture_logs(cmd_log)

    assert f"App name = {app_name}" in output
    assert "Enclave size = 4096M" in output
    assert f"Common name = {host}" in output
    assert f"Port = {port}" in output
    assert "Healthcheck = /health" in output
    assert "Status = waiting secret keys" in output


@pytest.mark.sgx
@pytest.mark.incremental
def test_evidence(workspace: Path, cmd_log: io.StringIO, app_name: str, pccs_url: str):
    """Test the `evidence` subcommand."""
    do_evidence(
        Namespace(
            **{
                "name": app_name,
                "pccs": pccs_url,
                "output": workspace,
            }
        )
    )

    output = capture_logs(cmd_log)
    try:
        pytest.evidence_path = Path(
            re.search(
                "The evidence file has been generated at: ([A-Za-z0-9/._-]+)", output
            ).group(1)
        )
    except AttributeError:
        print(output)
        assert False

    assert pytest.evidence_path.exists()


@pytest.mark.sgx
@pytest.mark.incremental
def test_verify(workspace: Path, cmd_log: io.StringIO):
    """Test the `verify` subcommand."""
    do_verify(
        Namespace(
            **{
                "package": pytest.package_path,
                "evidence": pytest.evidence_path,
                "output": workspace,
            }
        )
    )

    output = capture_logs(cmd_log)

    assert "Verification successful" in output

    try:
        pytest.ratls_cert = Path(
            re.search(
                "The RA-TLS certificate has been saved at: ([A-Za-z0-9/._-]+)", output
            ).group(1)
        )
    except AttributeError:
        print(output)
        assert False

    assert pytest.ratls_cert.exists()


@pytest.mark.sgx
@pytest.mark.incremental
def test_seal(workspace: Path, cmd_log: io.StringIO):
    """Test the `seal` subcommand."""
    input_path = pytest.app_path / "secrets_to_seal.json"

    do_seal(
        Namespace(
            **{
                "input": input_path,
                "receiver_enclave": pytest.ratls_cert,
                "receiver_public_key": None,
                "output": workspace / f"{input_path.name}.seal",
            }
        )
    )

    pytest.sealed_secrets = workspace / f"{input_path.name}.seal"

    assert pytest.sealed_secrets.exists()


@pytest.mark.sgx
@pytest.mark.incremental
def test_unseal(workspace: Path, cmd_log: io.StringIO):
    """Test the `unseal` subcommand."""
    input_path = pytest.app_path / "message.bin"
    input_path.write_bytes(b"Hello World!")

    output_path = workspace / f"{input_path.name}.seal"

    receiver_priv_key = X25519PrivateKey.generate()
    receiver_sk_path = workspace / "receiver_priv_key.bin"
    receiver_sk_path.write_bytes(
        receiver_priv_key.private_bytes(
            encoding=Encoding.Raw,
            format=PrivateFormat.Raw,
            encryption_algorithm=NoEncryption(),
        )
    )

    receiver_pub_key = receiver_priv_key.public_key()
    receiver_pk_path = workspace / "receiver_pub_key.bin"
    receiver_pk_path.write_bytes(
        receiver_pub_key.public_bytes(encoding=Encoding.Raw, format=PublicFormat.Raw)
    )

    do_seal(
        Namespace(
            **{
                "input": input_path,
                "receiver_enclave": None,
                "receiver_public_key": receiver_pk_path,
                "output": output_path,
            }
        )
    )
    do_unseal(
        Namespace(
            **{
                "input": output_path,
                "private_key": receiver_sk_path,
                "output": workspace / input_path.name,
            }
        )
    )

    assert (workspace / input_path.name).exists()


@pytest.mark.sgx
@pytest.mark.incremental
def test_run(cmd_log: io.StringIO, app_name: str):
    """Test the `run` subcommand."""
    do_run(
        Namespace(
            **{
                "name": app_name,
                "key": pytest.key_path,
                "timeout": 5,
                "secrets": pytest.app_path / "secrets.json",
                "sealed_secrets": pytest.sealed_secrets,
            }
        )
    )

    assert "Application ready!" in capture_logs(cmd_log)


@pytest.mark.sgx
@pytest.mark.incremental
def test_test(
    app_name: str,
    cmd_log: io.StringIO,
):
    """Test the `test` subcommand."""
    do_test(
        Namespace(
            **{
                "name": app_name,
                "test": pytest.app_path / "tests",
                "config": pytest.app_path / "config.toml",
            }
        )
    )

    assert "Tests successful" in capture_logs(cmd_log)


@pytest.mark.sgx
@pytest.mark.incremental
def test_encrypt(workspace: Path, port: int, host: str):
    """Test the `encrypt` subcommand."""
    plain_file_path = workspace / "test"
    plain_file_path.write_bytes(b"Hello World!")

    enc_file_path = workspace / "test.enc"

    key_path = workspace / "key.bin"
    key_path.write_bytes(base64.urlsafe_b64encode(os.urandom(32)))

    do_encrypt(
        Namespace(
            **{
                "input": plain_file_path,
                "key": key_path,
                "output": enc_file_path,
            }
        )
    )

    expected_plain_file_path = workspace / "test.plain"

    do_decrypt(
        Namespace(
            **{
                "input": enc_file_path,
                "key": key_path,
                "output": expected_plain_file_path,
            }
        )
    )

    assert expected_plain_file_path.exists()
    assert expected_plain_file_path.read_bytes() == plain_file_path.read_bytes()


@pytest.mark.sgx
@pytest.mark.incremental
def test_decrypt_secrets_json(workspace: Path, port: int, host: str):
    """Test the `decrypt` subcommand with the secret json file."""
    response = requests.get(
        url=f"https://{host}:{port}/result/secrets",
        verify=pytest.ratls_cert,
        timeout=5,
    )

    assert response.status_code == 200

    enc_file_path = workspace / "result.enc"
    enc_file_path.write_bytes(response.content)

    key_path = workspace / "key.bin"
    salt: bytes = base64.urlsafe_b64decode(
        b"NICFe3qjykoE_DCSng22uzd6Pks3P2HHnaF-y9G18qo="
    )
    password: bytes = "qwerty".encode("utf-8")
    kdf = Scrypt(salt=salt, length=32, n=2**14, r=8, p=1)
    key_path.write_bytes(base64.urlsafe_b64encode(kdf.derive(password)))

    output_path = workspace / "result.plain"

    do_decrypt(
        Namespace(
            **{
                "input": enc_file_path,
                "key": key_path,
                "output": output_path,
            }
        )
    )

    assert output_path.exists()
    assert output_path.read_text() == "message using password and salt from SECRETS"


@pytest.mark.sgx
@pytest.mark.incremental
def test_decrypt_sealed_secrets_json(workspace: Path, port: int, host: str):
    """Test the `decrypt` subcommand with the sealed secret json file."""
    response = requests.get(
        url=f"https://{host}:{port}/result/sealed_secrets",
        verify=pytest.ratls_cert,
        timeout=5,
    )

    assert response.status_code == 200

    enc_file_path = workspace / "result2.enc"
    enc_file_path.write_bytes(response.content)

    key_path = workspace / "key2.bin"
    key_path.write_bytes(b"AXIKImqLJa5SZms7o6mb_nL1QLro_GDNcpIJQ71CgMk=")

    output_path = workspace / "result2.plain"

    do_decrypt(
        Namespace(
            **{
                "input": enc_file_path,
                "key": key_path,
                "output": output_path,
            }
        )
    )

    assert output_path.exists()
    assert output_path.read_text() == "message using result_sk from SEALED_SECRETS"


@pytest.mark.sgx
@pytest.mark.incremental
def test_status(cmd_log: io.StringIO, app_name: str, port: int, host: str):
    """Test the `status` subcommand."""
    do_status(
        Namespace(
            **{
                "name": app_name,
            }
        )
    )

    output = capture_logs(cmd_log)

    assert f"App name = {app_name}" in output
    assert "Enclave size = 4096M" in output
    assert f"Common name = {host}" in output
    assert f"Port = {port}" in output
    assert "Healthcheck = /health" in output
    assert "Status = running" in output


@pytest.mark.sgx
@pytest.mark.incremental
def test_list(cmd_log: io.StringIO, app_name: str):
    """Test the `list` subcommand."""
    do_list(Namespace())

    output = capture_logs(cmd_log)

    assert f"running   | {app_name}" in output


@pytest.mark.sgx
@pytest.mark.incremental
def test_stop(cmd_log: io.StringIO, app_name: str):
    """Test the `stop` subcommand."""
    do_stop(Namespace(**{"name": app_name, "remove": False}))

    output = capture_logs(cmd_log)

    assert f"Docker '{app_name}' has been stopped!" in output
    assert f"Docker '{app_name}' has been removed!" not in output

    do_status(
        Namespace(
            **{
                "name": app_name,
            }
        )
    )

    output = capture_logs(cmd_log)

    assert "Status = exited" in output

    do_list(Namespace())

    output = capture_logs(cmd_log)

    assert f"exited   | {app_name}" in output


@pytest.mark.sgx
@pytest.mark.incremental
def test_restart(cmd_log: io.StringIO, app_name: str):
    """Test the `restart` subcommand."""
    do_restart(Namespace(**{"name": app_name}))

    output = capture_logs(cmd_log)

    assert f"Docker '{app_name}' is now restarted!" in output

    time.sleep(10)

    do_status(
        Namespace(
            **{
                "name": app_name,
            }
        )
    )

    output = capture_logs(cmd_log)

    assert "Status = running" in output

    do_list(Namespace())

    output = capture_logs(cmd_log)

    assert f"running   | {app_name}" in output


@pytest.mark.sgx
@pytest.mark.incremental
def test_remove(cmd_log: io.StringIO, app_name: str):
    """Test the `stop` subcommand with removing."""
    do_stop(Namespace(**{"name": app_name, "remove": True}))

    output = capture_logs(cmd_log)

    assert f"Docker '{app_name}' has been stopped!" in output
    assert f"Docker '{app_name}' has been removed!" in output

    with pytest.raises(Exception):
        do_status(
            Namespace(
                **{
                    "name": app_name,
                }
            )
        )

    do_list(Namespace())

    output = capture_logs(cmd_log)

    assert f"{app_name}" not in output


@pytest.mark.sgx
@pytest.mark.incremental
def test_plaintext(
    workspace: Path,
    cmd_log: io.StringIO,
    app_name: str,
    port2: int,
    host: str,
    signer_key: Path,
    pccs_url: str,
):
    """Test the process subcommand without encryption."""
    do_package(
        Namespace(
            **{
                "project": None,
                "code": pytest.app_path / "src",
                "config": pytest.app_path / "config.toml",
                "dockerfile": pytest.app_path / "Dockerfile",
                "test": pytest.app_path / "tests",
                "encrypt": False,  # We do not encrypt here
                "output": workspace,
            }
        )
    )

    # Check the tar generation
    output = capture_logs(cmd_log)
    try:
        assert "Your code secret key has been saved at" not in output

        pytest.package_path = Path(
            re.search(
                "Your package is now ready to be shared: ([A-Za-z0-9/._-]+)", output
            ).group(1)
        )
    except AttributeError:
        print(output)
        assert False

    assert pytest.package_path.exists()

    do_spawn(
        Namespace(
            **{
                "name": app_name,
                "package": pytest.package_path,
                "host": host,
                "subject": f"CN={host},O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
                "san": host,
                "days": 2,
                # We use `port2` because we do not manage when
                # docker releases the free previous port
                "port": port2,
                "size": 4096,
                "timeout": 5,
                "signer_key": signer_key,
                "pccs": pccs_url,
                "output": workspace,
            }
        )
    )

    output = capture_logs(cmd_log)
    try:
        pytest.evidence_path = Path(
            re.search(
                "The evidence file has been generated at: ([A-Za-z0-9/._-]+)", output
            ).group(1)
        )
    except AttributeError:
        print(output)
        assert False

    assert pytest.evidence_path.exists()

    do_run(
        Namespace(
            **{
                "name": app_name,
                "key": None,
                "timeout": 5,
                "secrets": None,
                "sealed_secrets": None,
            }
        )
    )

    assert "Application ready!" in capture_logs(cmd_log)

    do_test(
        Namespace(
            **{
                "name": app_name,
                "test": pytest.app_path / "tests",
                "config": pytest.app_path / "config.toml",
            }
        )
    )

    assert "Tests successful" in capture_logs(cmd_log)

    do_stop(Namespace(**{"name": app_name, "remove": True}))

    output = capture_logs(cmd_log)

    assert f"Docker '{app_name}' has been stopped!" in output
    assert f"Docker '{app_name}' has been removed!" in output


@pytest.mark.sgx
@pytest.mark.incremental
def test_plaintext_project(
    workspace: Path,
    cmd_log: io.StringIO,
    app_name: str,
    port3: int,
    host: str,
    signer_key: Path,
    pccs_url: str,
):
    """Test the process subcommand without encryption by specifying a project directory."""
    do_package(
        Namespace(
            **{
                "project": pytest.app_path,
                "code": None,
                "config": None,
                "dockerfile": None,
                "test": None,
                "encrypt": False,  # We do not encrypt here
                "output": workspace,
            }
        )
    )

    # Check the tar generation
    output = capture_logs(cmd_log)
    try:
        assert "Your code secret key has been saved at" not in output

        pytest.package_path = Path(
            re.search(
                "Your package is now ready to be shared: ([A-Za-z0-9/._-]+)", output
            ).group(1)
        )
    except AttributeError:
        print(output)
        assert False

    assert pytest.package_path.exists()

    do_spawn(
        Namespace(
            **{
                "name": app_name,
                "pccs": pccs_url,
                "package": pytest.package_path,
                "host": host,
                "subject": f"CN={host},O=MyApp Company,C=FR,L=Paris,ST=Ile-de-France",
                "san": host,
                "days": 2,
                # We use `port3` because we do not manage when
                # docker releases the free previous port
                "port": port3,
                "timeout": 5,
                "size": 4096,
                "signer_key": signer_key,
                "output": workspace,
            }
        )
    )

    output = capture_logs(cmd_log)
    try:
        pytest.evidence_path = Path(
            re.search(
                "The evidence file has been generated at: ([A-Za-z0-9/._-]+)", output
            ).group(1)
        )
    except AttributeError:
        print(output)
        assert False

    assert pytest.evidence_path.exists()

    do_run(
        Namespace(
            **{
                "name": app_name,
                "key": None,
                "timeout": 5,
                "secrets": None,
                "sealed_secrets": None,
            }
        )
    )

    assert "Application ready!" in capture_logs(cmd_log)

    do_test(
        Namespace(
            **{
                "name": app_name,
                "test": pytest.app_path / "tests",
                "config": pytest.app_path / "config.toml",
            }
        )
    )

    assert "Tests successful" in capture_logs(cmd_log)

    do_stop(Namespace(**{"name": app_name, "remove": True}))

    output = capture_logs(cmd_log)

    assert f"Docker '{app_name}' has been stopped!" in output
    assert f"Docker '{app_name}' has been removed!" in output



"""Test ignore_file.py."""

from pathlib import Path

from cenclave.core.ignore_file import IgnoreFile


def test_ignore_file():
    """Test `IgnoreFile`."""
    ignored_files = IgnoreFile.parse(Path(__file__).parent / "data")

    expected_patterns = {
        "config.toml",
        "secrets.json",
        ".cenclaveignore",
        "__pycache__/",
        ".git*",
        ".vscode/",
        ".idea/",
        "*.egg-info/",
        "*.egg",
        ".venv",
        ".env",
    }

    assert ignored_files == expected_patterns



"""cenclave.command.sgx_operator.test module."""

import os
import subprocess
import sys
from pathlib import Path

from cenclave.command.helpers import get_client_docker, get_running_app_container
from cenclave.core.bootstrap import is_waiting_for_secrets
from cenclave.core.conf import AppConf, AppConfParsingOption
from cenclave.core.sgx_docker import SgxDockerConfig
from cenclave.error import AppContainerBadState
from cenclave.log import LOGGER as LOG


def add_subparser(subparsers):
    """Define the subcommand."""
    parser = subparsers.add_parser("test", help="Test a deployed web app")

    parser.add_argument(
        "name",
        type=str,
        help="the name of the application",
    )

    parser.add_argument(
        "--test",
        type=Path,
        required=True,
        help="the path of the test directory extracted from the tarball package",
    )

    parser.add_argument(
        "--config",
        type=Path,
        required=True,
        help="the conf path extracted from the tarball package",
    )

    parser.set_defaults(func=run)


def run(args) -> None:
    """Run the subcommand."""
    client = get_client_docker()
    container = get_running_app_container(client, args.name)

    docker = SgxDockerConfig.load(container.attrs, container.labels)

    if is_waiting_for_secrets(f"https://{docker.host}:{docker.port}"):
        raise AppContainerBadState(
            "Your application is waiting for secrets and can't be tested right now."
        )

    code_config = AppConf.load(args.config, option=AppConfParsingOption.SkipCloud)

    for package in code_config.tests_requirements:
        subprocess.check_call([sys.executable, "-m", "pip", "install", package])

    try:
        env = dict(os.environ)

        if "TEST_REMOTE_URL" not in os.environ:
            env["TEST_REMOTE_URL"] = f"https://{docker.host}:{docker.port}"

        subprocess.check_call(
            [sys.executable, "-m", code_config.tests_cmd],
            cwd=args.test,
            env=env,
        )

        LOG.info("Tests successful")
    except subprocess.CalledProcessError:
        LOG.error("Tests failed!")
